from __future__ import absolute_import
from __future__ import annotations

import inspect
import logging
import os
from copy import deepcopy
from dataclasses import dataclass, field
from enum import IntEnum
from functools import wraps, lru_cache, cached_property
from itertools import combinations_with_replacement

import matplotlib.pyplot as plt
import numpy as np
import rustworkx as rx
import rustworkx.visit
import rustworkx.visualization
from lxml import etree
from random_events.utils import SubclassJSONSerializer
from rustworkx import NoEdgeBetweenNodes
from typing_extensions import (
    Dict,
    Tuple,
    Optional,
    TypeVar,
    Union,
    Callable,
    Any,
    Iterable,
)
from typing_extensions import List
from typing_extensions import Type, Set

from .callbacks.callback import StateChangeCallback, ModelChangeCallback
from .collision_checking.collision_detector import CollisionDetector
from .collision_checking.trimesh_collision_detector import TrimeshCollisionDetector
from .datastructures.prefixed_name import PrefixedName
from .datastructures.types import NpMatrix4x4
from .exceptions import (
    DuplicateSemanticAnnotationError,
    AddingAnExistingSemanticAnnotationError,
    SemanticAnnotationNotFoundError,
    AlreadyBelongsToAWorldError,
    DuplicateKinematicStructureEntityError,
    MissingWorldModificationContextError,
)
from .robots.abstract_robot import AbstractRobot
from .spatial_computations.forward_kinematics import ForwardKinematicsVisitor
from .spatial_computations.ik_solver import InverseKinematicsSolver
from .spatial_computations.raytracer import RayTracer
from .spatial_types import spatial_types as cas
from .spatial_types.derivatives import Derivatives
from .utils import IDGenerator, copy_lru_cache
from .world_description.connections import (
    ActiveConnection,
    PassiveConnection,
    FixedConnection,
    Connection6DoF,
    ActiveConnection1DOF,
)
from .world_description.connections import HasUpdateState
from .world_description.degree_of_freedom import DegreeOfFreedom
from .world_description.world_entity import (
    Connection,
    SemanticAnnotation,
    KinematicStructureEntity,
    Region,
    GenericKinematicStructureEntity,
    GenericConnection,
    CollisionCheckingConfig,
    Body,
)
from .world_description.world_modification import (
    RemoveSemanticAnnotationModification,
)
from .world_description.world_modification import (
    WorldModelModification,
    AddDegreeOfFreedomModification,
    RemoveDegreeOfFreedomModification,
    AddKinematicStructureEntityModification,
    AddConnectionModification,
    RemoveBodyModification,
    RemoveConnectionModification,
    WorldModelModificationBlock,
    SetDofHasHardwareInterface,
    AddSemanticAnnotationModification,
)
from .world_description.world_state import WorldState

logger = logging.getLogger(__name__)

id_generator = IDGenerator()

GenericSemanticAnnotation = TypeVar(
    "GenericSemanticAnnotation", bound=SemanticAnnotation
)

FunctionStack = List[Tuple[Callable, Dict[str, Any]]]


class PlotAlignment(IntEnum):
    HORIZONTAL = 0
    VERTICAL = 1


class ResetStateContextManager:
    """
    A context manager for resetting the state of a given `World` instance.

    This class is designed to allow operations to be performed on a `World`
    object, ensuring that its state can be safely returned to its previous
    condition upon leaving the context. If no exceptions occur within the
    context, the original state of the `World` instance is restored, and the
    state change is notified.
    """

    def __init__(self, world: World):
        self.world = world

    def __enter__(self) -> None:
        self.state = deepcopy(self.world.state)

    def __exit__(
        self,
        exc_type: Optional[type],
        exc_val: Optional[Exception],
        exc_tb: Optional[type],
    ) -> None:
        if exc_type is None:
            self.world.state = self.state
            self.world.notify_state_change()


class WorldModelUpdateContextManager:
    """
    Context manager for updating the state of a given `World` instance.
    This class manages that updates to the world within the context of this class only trigger recomputations after all
    desired updates have been performed.
    """

    first: bool = True
    """
    First time flag.
    """

    modification: Callable = None
    """
    Modification function.
    """

    arguments: Dict[str, Any] = None
    """
    Arguments of the modification function.
    """

    def __init__(self, world: World):
        self.world = world

    def __enter__(self):
        if self.world.world_is_being_modified:
            self.first = False
        self.world.world_is_being_modified = True

        if self.first:
            self.world._current_model_modification_block = WorldModelModificationBlock()

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.first:
            self.world._model_modification_blocks.append(
                self.world._current_model_modification_block
            )
            self.world._current_model_modification_block = None
            if exc_type is None:
                self.world._notify_model_change()
            self.world.world_is_being_modified = False


class AtomicWorldModificationNotAtomic(Exception):
    """
    Exception raised when atomic world modifications are overlapping.
    If this exception is raised, it means that somewhere in the code a function decorated with @atomic_world_modification
    triggered another function decorated with it. This must not happen ever!
    """


def atomic_world_modification(
    func=None, modification: Type[WorldModelModification] = None
):
    """
    Decorator for ensuring atomicity in world modification operations.

    This decorator ensures that no other atomic world modification is in progress when the decorated function is executed.
    It records the function call along with its arguments for potential replay or tracking purposes.
    If an operation is attempted when the world is locked, it raises an appropriate exception.

    Raises:
        AtomicWorldModificationNotAtomic: If the world is already locked during the execution of another atomic operation.
    """

    def _decorate(func):

        sig = inspect.signature(func)

        @wraps(func)
        def wrapper(self: World, *args, **kwargs):
            if self._atomic_modification_is_being_executed:
                raise AtomicWorldModificationNotAtomic(f"World {self} is locked.")
            self._atomic_modification_is_being_executed = True

            # bind args and kwargs
            bound = sig.bind_partial(
                self, *args, **kwargs
            )  # use bind() if you require all args
            bound.apply_defaults()  # fill in default values

            # record function call
            # Build a dict with all arguments (including positional), excluding 'self'
            bound_args = dict(bound.arguments)
            bound_args.pop("self", None)
            if self._current_model_modification_block is None:
                raise MissingWorldModificationContextError(func)
            self._current_model_modification_block.append(
                modification.from_kwargs(bound_args)
            )

            result = func(self, *args, **kwargs)

            self._atomic_modification_is_being_executed = False
            return result

        return wrapper

    if func is None:
        return _decorate

    return _decorate(func)


@dataclass
class World:
    """
    A class representing the world.
    The world manages a set of kinematic structure entities and connections represented as a tree-like graph.
    The nodes represent kinematic structure entities in the world, and the edges represent joins between them.
    """

    kinematic_structure: rx.PyDAG[KinematicStructureEntity] = field(
        default_factory=lambda: rx.PyDAG(multigraph=False), kw_only=True, repr=False
    )
    """
    The kinematic structure of the world.
    The kinematic structure is a tree shaped directed graph where the nodes represent kinematic structure entities
     in the world, and the edges represent connections between them.
    """

    semantic_annotations: List[SemanticAnnotation] = field(
        default_factory=list, repr=False
    )
    """
    All semantic annotations the world is aware of.
    """

    degrees_of_freedom: List[DegreeOfFreedom] = field(default_factory=list)
    """
    All degrees of freedom in the world.
    """

    state: WorldState = field(default_factory=WorldState)
    """
    2d array where rows are derivatives and columns are dof values for that derivative.
    """

    _model_version: int = 0
    """
    The version of the model. This increases whenever a change to the kinematic model is made. Mostly triggered
    by adding/removing bodies and connections.
    """

    _state_version: int = 0
    """
    The version of the state. This increases whenever a change to the state of the kinematic model is made. 
    Mostly triggered by updating connection values.
    """

    world_is_being_modified: bool = False
    """
    Is set to True, when a world.modify_world context is used.
    """

    name: Optional[str] = None
    """
    Name of the world. May act as default namespace for all bodies and semantic annotations in the world which do not have a prefix.
    """

    state_change_callbacks: List[StateChangeCallback] = field(
        default_factory=list, repr=False
    )
    """
    Callbacks to be called when the state of the world changes.
    """

    model_change_callbacks: List[ModelChangeCallback] = field(
        default_factory=list, repr=False
    )
    """
    Callbacks to be called when the model of the world changes.
    """

    _model_modification_blocks: List[WorldModelModificationBlock] = field(
        default_factory=list, repr=False, init=False
    )
    """
    All atomic modifications applied to the world. Tracked by @atomic_world_modification.
    The field itself is a list of lists. The outer lists indicates when to trigger the model/state change callbacks.
    The inner list is a block of modifications where change callbacks must not be called in between.
    """

    _current_model_modification_block: Optional[WorldModelModificationBlock] = field(
        default=None, repr=False, init=False
    )
    """
    The current modification block called within one context of @atomic_world_modification.
    """

    _atomic_modification_is_being_executed: bool = field(init=False, default=False)
    """
    Flag that indicates if an atomic world operation is currently being executed.
    See `atomic_world_modification` for more information.
    """

    _disabled_collision_pairs: Set[
        Tuple[KinematicStructureEntity, KinematicStructureEntity]
    ] = field(default_factory=set, repr=False)
    """
    Collisions for these Body pairs is disabled.f
    """

    _temp_disabled_collision_pairs: Set[
        Tuple[KinematicStructureEntity, KinematicStructureEntity]
    ] = field(default_factory=set, repr=False)
    """
    A set of Body pairs for which collisions are temporarily disabled.
    """

    def reset_temporary_collision_config(self):
        self._temp_disabled_collision_pairs = set()
        for body in self.bodies:
            if body.has_collision():
                body.reset_temporary_collision_config()

    @property
    def root(self):
        return self._root(self._model_version)

    @lru_cache()
    def _root(self, model_version) -> Optional[KinematicStructureEntity]:
        """
        The root of the world is the unique node with in-degree 0.

        :param model_version: Version of the model used to manage the cache of return values.
        :return: The root of the world.
        """
        if not self.kinematic_structure_entities:
            return None
        possible_roots = [
            node
            for node in self.kinematic_structure_entities
            if self.kinematic_structure.in_degree(node.index) == 0
        ]
        if len(possible_roots) == 1:
            return possible_roots[0]
        elif len(possible_roots) > 1:
            raise ValueError(
                f"More than one root found. Possible roots are {possible_roots}"
            )
        else:
            raise ValueError(f"No root found.")

    def __hash__(self):
        return hash(id(self))

    @property
    def active_degrees_of_freedom(self) -> Set[DegreeOfFreedom]:
        dofs = set()
        for connection in self.connections:
            if isinstance(connection, ActiveConnection):
                dofs.update(set(connection.active_dofs))
        return dofs

    @property
    def passive_degrees_of_freedom(self) -> Set[DegreeOfFreedom]:
        dofs = set()
        for connection in self.connections:
            if isinstance(connection, PassiveConnection):
                dofs.update(set(connection.passive_dofs))
        return dofs

    @cached_property
    def collision_detector(self) -> CollisionDetector:
        """
        A collision detector for the world.
        :return: A collision detector for the world.
        """
        return TrimeshCollisionDetector(self)

    @cached_property
    def ray_tracer(self) -> RayTracer:
        """
        A ray tracer for the world.
        :return: A ray tracer for the world.
        """
        return RayTracer(self)

    def validate(self) -> bool:
        """
        Validate the world.

        The world must be a tree.
        :return: True if the world is valid, raises an AssertionError otherwise.
        """
        if self.empty:
            return True
        assert len(self.kinematic_structure_entities) == (len(self.connections) + 1)
        assert rx.is_weakly_connected(self.kinematic_structure)
        actual_dofs = set()
        for connection in self.connections:
            actual_dofs.update(connection.dofs)
        assert actual_dofs == set(
            self.degrees_of_freedom
        ), "self.degrees_of_freedom does not match the actual dofs used in connections. Did you forget to call deleted_orphaned_dof()?"
        return True

    @atomic_world_modification(modification=AddDegreeOfFreedomModification)
    def _add_degree_of_freedom(self, dof: DegreeOfFreedom) -> None:
        """
        Adds a degree of freedom to the current system and initializes its state.

        This method modifies the internal state of the system by adding a new
        degree of freedom (DOF). It sets the initial position of the DOF based
        on its configured lower and upper position limits, ensuring it respects
        both constraints. The DOF is then added to the list of degrees of freedom
        in the system.

        :param dof: The degree of freedom to be added to the system.
        :return: None
        """
        dof._world = self
        dof.create_and_register_symbols()

        initial_position = 0
        lower_limit = dof.lower_limits.position
        if lower_limit is not None:
            initial_position = max(lower_limit, initial_position)
        upper_limit = dof.upper_limits.position
        if upper_limit is not None:
            initial_position = min(upper_limit, initial_position)
        self.state[dof.name].position = initial_position
        self.degrees_of_freedom.append(dof)

    def add_degree_of_freedom(self, dof: DegreeOfFreedom) -> None:
        """
        Adds degree of freedom in the world.
        This is used to register DoFs that are not created by the world, but are part of the world model.
        :param dof: The degree of freedom to register.
        """
        if dof._world is self and dof in self.degrees_of_freedom:
            return
        if dof._world is not None:
            raise AlreadyBelongsToAWorldError(
                world=dof._world, type_trying_to_add=DegreeOfFreedom
            )
        self._add_degree_of_freedom(dof)

    @atomic_world_modification(modification=RemoveDegreeOfFreedomModification)
    def _remove_degree_of_freedom(self, dof: DegreeOfFreedom) -> None:
        dof._world = None
        self.degrees_of_freedom.remove(dof)
        del self.state[dof.name]

    def remove_degree_of_freedom(self, dof: DegreeOfFreedom) -> None:
        if dof._world is self:
            self._remove_degree_of_freedom(dof)
        else:
            logger.debug("Trying to remove an dof that is not part of this world.")

    def modify_world(self) -> WorldModelUpdateContextManager:
        return WorldModelUpdateContextManager(self)

    def reset_state_context(self) -> ResetStateContextManager:
        return ResetStateContextManager(self)

    def clear_all_lru_caches(self):
        for method_name in dir(self):
            try:
                method = getattr(type(self), method_name)
                if hasattr(method, "cache_clear") and callable(method.cache_clear):
                    method.cache_clear()
            except AttributeError:
                # Skip attributes that can't be accessed
                pass

    def notify_state_change(self) -> None:
        """
        If you have changed the state of the world, call this function to trigger necessary events and increase
        the state version.
        """
        # self.compute_fk.cache_clear()
        # self.compute_fk_with_collision_offset_np.cache_clear()
        if not self.empty:
            self._recompute_forward_kinematics()
        self._state_version += 1
        for callback in self.state_change_callbacks:
            callback.notify()

    def _notify_model_change(self) -> None:
        """
        Notifies the system of a model change and updates necessary states, caches,
        and forward kinematics expressions while also triggering registered callbacks
        for model changes.
        """
        # if not self.world_is_being_modified:
        self.clear_all_lru_caches()
        self.compile_forward_kinematics_expressions()
        self.notify_state_change()
        self._model_version += 1

        for callback in self.model_change_callbacks:
            callback.notify()

        for callback in self.state_change_callbacks:
            callback.update_previous_world_state()

        self.validate()
        self.disable_non_robot_collisions()
        self.disable_collisions_for_adjacent_bodies()

    def delete_orphaned_dofs(self):
        actual_dofs = set()
        for connection in self.connections:
            actual_dofs.update(connection.dofs)
        self.degrees_of_freedom = list(actual_dofs)

    def get_kinematic_structure_entity_by_type(
        self, entity_type: Type[GenericKinematicStructureEntity]
    ) -> List[GenericKinematicStructureEntity]:
        """
        Retrieves all kinematic structure entities of a specific type from the world.

        :param entity_type: The class (type) of the kinematic structure entities to search for.
        :return: A list of `KinematicStructureEntity` objects that match the given type.
        """
        return [
            entity
            for entity in self.kinematic_structure_entities
            if isinstance(entity, entity_type)
        ]

    @property
    def kinematic_structure_entities(self) -> List[KinematicStructureEntity]:
        """
        :return: A list of all bodies in the world.
        """
        return list(self.kinematic_structure.nodes())

    @property
    def regions(self) -> List[Region]:
        """
        :return: A list of all regions in the world.
        """
        return self.get_kinematic_structure_entity_by_type(Region)

    @property
    def bodies(self) -> List[Body]:
        """
        :return: A list of all bodies in the world.
        """
        return self.get_kinematic_structure_entity_by_type(Body)

    @property
    def connections(self) -> List[Connection]:
        """
        :return: A list of all connections in the world.
        """
        return list(self.kinematic_structure.edges())

    @atomic_world_modification(modification=AddKinematicStructureEntityModification)
    def _add_kinematic_structure_entity(
        self, kinematic_structure_entity: KinematicStructureEntity
    ) -> int:
        """
        Add a kinematic_structure_entity to the world.
        Do not call this function directly, use add_kinematic_structure_entity instead.

        :param kinematic_structure_entity: The kinematic_structure_entity to add.
        :return: The index of the added kinematic_structure_entity.
        """
        index = kinematic_structure_entity.index = self.kinematic_structure.add_node(
            kinematic_structure_entity
        )
        kinematic_structure_entity._world = self
        return index

    def add_kinematic_structure_entity(
        self,
        kinematic_structure_entity: KinematicStructureEntity,
        handle_duplicates: bool = False,
    ) -> Optional[int]:
        """
        Add a kinematic_structure_entity to the world if it does not exist already.

        :param kinematic_structure_entity: The kinematic_structure_entity to add.
        :param handle_duplicates: If True, the kinematic_structure_entity will be added under a different name, if
        the name already exists. If False, an error will be raised. Default is False.
        :return: The index of the added kinematic_structure_entity.
        """
        logger.info(
            f"Trying to add kinematic_structure_entity with name {kinematic_structure_entity.name}"
        )
        if (
            kinematic_structure_entity._world is self
            and kinematic_structure_entity.index is not None
        ):
            logger.info(
                f"Skipping since add kinematic_structure_entity already exists."
            )
            return None
        elif (
            kinematic_structure_entity._world is not None
            and kinematic_structure_entity._world is not self
        ):
            raise AlreadyBelongsToAWorldError(
                world=kinematic_structure_entity._world,
                type_trying_to_add=KinematicStructureEntity,
            )
        elif kinematic_structure_entity.name in [
            ke.name for ke in self.kinematic_structure_entities
        ]:
            if not handle_duplicates:
                raise DuplicateKinematicStructureEntityError(
                    [kinematic_structure_entity.name]
                )
            kinematic_structure_entity.name.name = (
                kinematic_structure_entity.name.name
                + f"_{id_generator(kinematic_structure_entity)}"
            )

        return self._add_kinematic_structure_entity(kinematic_structure_entity)

    def add_body(
        self, body: KinematicStructureEntity, handle_duplicates: bool = False
    ) -> Optional[int]:
        return self.add_kinematic_structure_entity(body, handle_duplicates)

    @atomic_world_modification(modification=AddConnectionModification)
    def _add_connection(self, connection: Connection):
        """
        Adds a connection to the kinematic structure.

        The method updates the connection instance to associate it with the current
        world instance and reflects the connection in the kinematic structure.
        Do not call this function directly, use add_connection instead.

        :param connection: The connection to be added to the kinematic structure.
        """
        connection._world = self
        self.kinematic_structure.add_edge(
            connection.parent.index, connection.child.index, connection
        )

    @atomic_world_modification(modification=SetDofHasHardwareInterface)
    def set_dofs_has_hardware_interface(
        self, dofs: Iterable[DegreeOfFreedom], value: bool
    ):
        """
        Sets whether the specified degrees of freedom (DOFs) have a hardware interface or not.

        This method allows controlling the presence of a hardware interface for multiple
        DOFs at once. The modification is atomic, ensuring that all DOFs are updated as
        a single operation and the state remains consistent. The method iterates through
        the given DOFs and updates their `has_hardware_interface` attribute to the provided
        value.

        :param dofs: An iterable collection of DegreeOfFreedom instances whose
                     `has_hardware_interface` attribute is to be updated.
        :param value: A boolean value indicating whether the DOFs should have a hardware
                      interface (True) or not (False).
        """
        for dof in dofs:
            dof.has_hardware_interface = value

    def add_connection(
        self,
        connection: Connection,
        handle_duplicates: bool = False,
    ) -> None:
        """
        Add a connection and the entities it connects to the world.

        :param connection: The connection to add.
        """
        connection.add_to_world(self)
        for dof in connection.dofs:
            if dof._world is None:
                self.add_degree_of_freedom(dof)
        self.add_kinematic_structure_entity(connection.parent, handle_duplicates)
        self.add_kinematic_structure_entity(connection.child, handle_duplicates)

        self._add_connection(connection)

    def add_semantic_annotation(
        self, semantic_annotation: SemanticAnnotation, exists_ok: bool = False
    ) -> None:
        """
        Adds a semantic annotation to the current list of semantic annotations if it doesn't already exist. Ensures
        that the `semantic_annotation` is associated with the current instance and maintains the
        integrity of unique semantic annotation names.

        :param semantic_annotation: The semantic annotation instance to be added. Its name must be unique within
            the current context.
        :param exists_ok: Whether to raise an error or not when a semantic annotation already exists.

        :raises AddingAnExistingSemanticAnnotationError: If exists_ok is False and a semantic annotation with the same name and type already exists.
        """
        try:
            self.get_semantic_annotation_by_name(semantic_annotation.name)
            if not exists_ok:
                raise AddingAnExistingSemanticAnnotationError(semantic_annotation)
        except SemanticAnnotationNotFoundError:
            self._add_semantic_annotation(semantic_annotation)

    @atomic_world_modification(modification=AddSemanticAnnotationModification)
    def _add_semantic_annotation(self, semantic_annotation: SemanticAnnotation):
        """
        The atomic method that adds a semantic annotation to the current list of semantic annotations.
        """
        semantic_annotation._world = self
        self.semantic_annotations.append(semantic_annotation)

    def remove_semantic_annotation(
        self, semantic_annotation: SemanticAnnotation
    ) -> None:
        """
        Removes a semantic annotation from the current list of semantic annotations if it exists.

        :param semantic_annotation: The semantic annotation instance to be removed.
        """
        try:
            existing_semantic_annotation = self.get_semantic_annotation_by_name(
                semantic_annotation.name
            )
            if existing_semantic_annotation == semantic_annotation:
                self._remove_semantic_annotation(semantic_annotation)
            else:
                raise ValueError(
                    "The provided semantic annotation instance does not match the existing semantic annotation with the same name."
                )
        except SemanticAnnotationNotFoundError:
            logger.debug(
                f"semantic annotation {semantic_annotation.name} not found in the world. No action taken."
            )

    @atomic_world_modification(modification=RemoveSemanticAnnotationModification)
    def _remove_semantic_annotation(self, semantic_annotation: SemanticAnnotation):
        """
        The atomic method that removes a semantic annotation from the current list of semantic annotations.
        """
        self.semantic_annotations.remove(semantic_annotation)
        semantic_annotation._world = None

    def get_connections_of_branch(
        self, root: KinematicStructureEntity
    ) -> List[Connection]:
        """
        Collect all connections that are below root in the tree.

        :param root: The root body of the branch
        :return: List of all connections in the subtree rooted at the given body
        """

        # Create a custom visitor to collect connections
        class ConnectionCollector(rustworkx.visit.DFSVisitor):
            def __init__(self, world: "World"):
                self.world = world
                self.connections = []

            def tree_edge(self, edge: Tuple[int, int, Connection]):
                """Called for each tree edge during DFS traversal"""
                self.connections.append(edge[2])  # edge[2] is the connection

        visitor = ConnectionCollector(self)
        rx.dfs_search(self.kinematic_structure, [root.index], visitor)

        return visitor.connections

    def get_kinematic_structure_entities_of_branch(
        self, root: KinematicStructureEntity
    ) -> List[KinematicStructureEntity]:
        """
        Collect all bodies that are below root in the tree.

        :param root: The root body of the branch
        :return: List of all bodies in the subtree rooted at the given body (including the root)
        """
        descendants_indices = rx.descendants(self.kinematic_structure, root.index)
        return [root] + [
            self.kinematic_structure[index] for index in descendants_indices
        ]

    def get_semantic_annotation_by_name(
        self, name: Union[str, PrefixedName]
    ) -> Optional[SemanticAnnotation]:
        """
        Retrieves a semantic annotation from the list of semantic annotation based on its name.
        If the input is of type `PrefixedName`, it checks whether the prefix is specified and looks for an
        exact match. Otherwise, it matches based on the name's string representation.
        If more than one body with the same name is found, an assertion error is raised.
        If no matching body is found, a `ValueError` is raised.

        :param name: The name of the semantic annotation to search for. Can be a string or a `PrefixedName` object.
        :return: The `SemanticAnnotation` object that matches the given name.
        :raises ValueError: If multiple or no semantic annotations with the specified name are found.
        :raises KeyError: If no semantic annotation is found.
        """
        if isinstance(name, PrefixedName):
            if name.prefix is not None:
                matches = [
                    semantic_annotation
                    for semantic_annotation in self.semantic_annotations
                    if semantic_annotation.name == name
                ]
            else:
                matches = [
                    semantic_annotation
                    for semantic_annotation in self.semantic_annotations
                    if semantic_annotation.name.name == name.name
                ]
        else:
            matches = [
                semantic_annotation
                for semantic_annotation in self.semantic_annotations
                if semantic_annotation.name.name == name
            ]
        if len(matches) > 1:
            raise DuplicateSemanticAnnotationError(matches)
        if matches:
            return matches[0]
        raise SemanticAnnotationNotFoundError(name)

    def get_world_state_symbols(self) -> List[cas.Symbol]:
        """
        Constructs and returns a list of symbols representing the state of the system. The state
        is defined in terms of positions, velocities, accelerations, and jerks for each degree
        of freedom specified in the current state.

        :raises KeyError: If a degree of freedom defined in the state does not exist in
            the `degrees_of_freedom`.
        :returns: A combined list of symbols corresponding to the positions, velocities,
            accelerations, and jerks for each degree of freedom in the state.
        """
        positions = [
            self.get_degree_of_freedom_by_name(v_name).symbols.position
            for v_name in self.state
        ]
        velocities = [
            self.get_degree_of_freedom_by_name(v_name).symbols.velocity
            for v_name in self.state
        ]
        accelerations = [
            self.get_degree_of_freedom_by_name(v_name).symbols.acceleration
            for v_name in self.state
        ]
        jerks = [
            self.get_degree_of_freedom_by_name(v_name).symbols.jerk
            for v_name in self.state
        ]
        return positions + velocities + accelerations + jerks

    def get_semantic_annotations_by_type(
        self, semantic_annotation_type: Type[GenericSemanticAnnotation]
    ) -> List[GenericSemanticAnnotation]:
        """
        Retrieves all semantic annotations of a specific type from the world.

        :param semantic_annotation_type: The class (type) of the semantic annotations to search for.
        :return: A list of `SemanticAnnotation` objects that match the given type.
        """
        return [
            semantic_annotation
            for semantic_annotation in self.semantic_annotations
            if isinstance(semantic_annotation, semantic_annotation_type)
        ]

    @atomic_world_modification(modification=RemoveBodyModification)
    def _remove_kinematic_structure_entity(
        self, kinematic_structure_entity: KinematicStructureEntity
    ) -> None:
        """
        Removes a kinematic_structure_entity from the world.

        Do not call this function directly, use `remove_kinematic_structure_entity` instead.

        :param kinematic_structure_entity: The kinematic_structure_entity to remove.
        """
        self.kinematic_structure.remove_node(kinematic_structure_entity.index)
        kinematic_structure_entity._world = None
        kinematic_structure_entity.index = None

    def remove_kinematic_structure_entity(
        self, kinematic_structure_entity: KinematicStructureEntity
    ) -> None:
        """
        Removes a kinematic_structure_entity from the world.

        :param kinematic_structure_entity: The kinematic_structure_entity to remove.
        """
        if (
            kinematic_structure_entity._world is self
            and kinematic_structure_entity.index is not None
        ):
            self._remove_kinematic_structure_entity(kinematic_structure_entity)
        else:
            logger.debug(
                "Trying to remove an kinematic_structure_entity that is not part of this world."
            )

    @atomic_world_modification(modification=RemoveConnectionModification)
    def _remove_connection(self, connection: Connection) -> None:
        try:
            self.kinematic_structure.remove_edge(
                connection.parent.index, connection.child.index
            )
        except NoEdgeBetweenNodes:
            pass
        connection._world = None
        connection.index = None

    def remove_connection(self, connection: Connection) -> None:
        """
        Removes a connection and deletes the corresponding degree of freedom, if it was only used by this connection.
        Might create disconnected entities, so make sure to add a new connection or delete the child kinematic_structure_entity.

        :param connection: The connection to be removed
        """
        remaining_dofs = set()
        for remaining_connection in self.connections:
            if remaining_connection == connection:
                continue
            remaining_dofs.update(remaining_connection.dofs)

        with self.modify_world():
            for dof in connection.dofs:
                if dof not in remaining_dofs:
                    self.remove_degree_of_freedom(dof)
            self._remove_connection(connection)

    def merge_world(
        self,
        other: World,
        root_connection: Connection = None,
        handle_duplicates: bool = False,
    ) -> None:
        """
        Merge a world into the existing one by merging degrees of freedom, states, connections, and bodies.
        This removes all bodies and connections from `other`.

        :param other: The world to be added.
        :param root_connection: If provided, this connection will be used to connect the two worlds. Otherwise, a new Connection6DoF will be created
        :param handle_duplicates: If True, bodies and semantic annotations with duplicate names will be renamed. If False, an error will be raised if duplicates are found.
        :return: None
        """
        assert other is not self, "Cannot merge a world with itself."

        with self.modify_world():
            old_state = deepcopy(other.state)
            self_root = self.root
            other_root = other.root
            with other.modify_world():
                for dof in other.degrees_of_freedom.copy():
                    other.remove_degree_of_freedom(dof)
                    self.add_degree_of_freedom(dof)
                for connection in other.connections:
                    other.remove_kinematic_structure_entity(connection.parent)
                    other.remove_kinematic_structure_entity(connection.child)
                    self.add_connection(connection, handle_duplicates=handle_duplicates)
                else:
                    other.remove_kinematic_structure_entity(other_root)
                    self.add_kinematic_structure_entity(other_root)
                for kinematic_structure_entity in other.kinematic_structure_entities:
                    if kinematic_structure_entity._world is not None:
                        other.remove_kinematic_structure_entity(
                            kinematic_structure_entity
                        )

                other_semantic_annotations = [
                    semantic_annotation
                    for semantic_annotation in other.semantic_annotations
                ]
                for semantic_annotation in other_semantic_annotations:
                    other.remove_semantic_annotation(semantic_annotation)
                    self.add_semantic_annotation(
                        semantic_annotation, exists_ok=handle_duplicates
                    )

            connection = root_connection
            if not connection and self_root:
                connection = Connection6DoF.create_with_dofs(
                    parent=self_root, child=other_root, world=self
                )

            if connection:
                self.add_connection(connection, handle_duplicates=handle_duplicates)

            for dof_name in old_state.keys():
                self.state[dof_name] = old_state[dof_name]

    def move_branch(
        self,
        branch_root: KinematicStructureEntity,
        new_parent: KinematicStructureEntity,
    ) -> None:
        """
        Destroys the connection between branch_root and its parent, and moves it to a new parent using a new connection
        of the same type. The pose of body with respect to root stays the same.

        :param branch_root: The root of the branch to be moved.
        :param new_parent: The new parent of the branch.
        """

        with self.modify_world():
            old_connection = branch_root.parent_connection
            if isinstance(old_connection, FixedConnection):
                new_parent_T_root = self.compute_forward_kinematics(
                    new_parent, branch_root
                )
                new_connection = FixedConnection(
                    parent=new_parent,
                    child=branch_root,
                    parent_T_connection_expression=new_parent_T_root,
                )
                self.add_connection(new_connection)
                self.remove_connection(old_connection)
            elif isinstance(old_connection, Connection6DoF):
                new_parent_T_root = self.compute_forward_kinematics(
                    new_parent, branch_root
                )
                new_connection = Connection6DoF.create_with_dofs(
                    parent=new_parent, child=branch_root, world=self
                )
                self.add_connection(new_connection)
                self.remove_connection(old_connection)
                new_connection.origin = new_parent_T_root
            else:
                raise ValueError(
                    f'Cannot move branch: "{branch_root.name}" is not connected with a FixedConnection'
                )

    def merge_world_at_pose(self, other: World, pose: cas.TransformationMatrix) -> None:
        """
        Merge another world into the existing one, creates a 6DoF connection between the root of this world and the root
        of the other world.
        :param other: The world to be added.
        :param pose: world_root_T_other_root, the pose of the other world's root with respect to the current world's root
        """
        with self.modify_world():
            root_connection = Connection6DoF.create_with_dofs(
                parent=self.root, child=other.root, world=self
            )
            self.merge_world(other, root_connection)
            root_connection.origin = pose

    def __str__(self):
        return f"{self.__class__.__name__} with {len(self.kinematic_structure_entities)} bodies."

    def get_connection(
        self, parent: KinematicStructureEntity, child: KinematicStructureEntity
    ) -> Connection:
        """
        Retrieves the connection between a parent and child kinematic_structure_entity in the kinematic structure.
        """
        return self.kinematic_structure.get_edge_data(parent.index, child.index)

    def get_connections_by_type(
        self, connection_type: Type[GenericConnection]
    ) -> List[GenericConnection]:
        return [c for c in self.connections if isinstance(c, connection_type)]

    def clear(self):
        """
        Clears all stored data and resets the state of the instance.
        """
        with self.modify_world():
            for body in list(self.bodies):
                self.remove_kinematic_structure_entity(body)

            self.semantic_annotations.clear()
            self.degrees_of_freedom.clear()
            self.state = WorldState()

    def get_kinematic_structure_entity_by_name(
        self, name: Union[str, PrefixedName]
    ) -> KinematicStructureEntity:
        """
        Retrieves a kinematic_structure_entity from the list of KinematicStructureEntites based on its name.
        If the input is of type `PrefixedName`, it checks whether the prefix is specified and looks for an
        exact match. Otherwise, it matches based on the name's string representation.
        If more than one kinematic_structure_entity with the same name is found, an assertion error is raised.
        If no matching kinematic_structure_entity is found, a `ValueError` is raised.

        :param name: The name of the kinematic_structure_entity to search for. Can be a string or a `PrefixedName` object.
        :return: The `KinematicStructureEntity` object that matches the given name.
        :raises ValueError: If multiple or no KinematicStructureEntities with the specified name are found.
        """
        if isinstance(name, PrefixedName):
            if name.prefix is not None:
                matches = [
                    entity
                    for entity in self.kinematic_structure_entities
                    if entity.name == name
                ]
            else:
                matches = [
                    entity
                    for entity in self.kinematic_structure_entities
                    if entity.name.name == name.name
                ]
        else:
            matches = [
                entity
                for entity in self.kinematic_structure_entities
                if entity.name.name == name
            ]
        if len(matches) > 1:
            raise ValueError(
                f"Multiple KinematicStructureEntities with name {name} found"
            )
        if matches:
            return matches[0]
        raise KeyError(f"KinematicStructureEntity with name {name} not found")

    def get_body_by_name(self, name: Union[str, PrefixedName]) -> Body:
        """
        Retrieves a Body from the list of bodies based on its name.
        If the input is of type `PrefixedName`, it checks whether the prefix is specified and looks for an
        exact match. Otherwise, it matches based on the name's string representation.
        If more than one body with the same name is found, an assertion error is raised.
        If no matching body is found, a `ValueError` is raised.

        :param name: The name of the body to search for. Can be a string or a `PrefixedName` object.
        :return: The `Body` object that matches the given name.
        :raises ValueError: If multiple or no bodies with the specified name are found.
        """
        if isinstance(name, PrefixedName):
            if name.prefix is not None:
                matches = [body for body in self.bodies if body.name == name]
            else:
                matches = [body for body in self.bodies if body.name.name == name.name]
        else:
            matches = [body for body in self.bodies if body.name.name == name]
        if len(matches) > 1:
            raise ValueError(f"Multiple bodies with name {name} found")
        if matches:
            return matches[0]
        raise KeyError(f"Body with name {name} not found")

    @lru_cache(maxsize=None)
    def get_degree_of_freedom_by_name(
        self, name: Union[str, PrefixedName]
    ) -> DegreeOfFreedom:
        """
        Retrieves a DegreeOfFreedom from the list of DegreeOfFreedom based on its name.
        If the input is of type `PrefixedName`, it checks whether the prefix is specified and looks for an
        exact match. Otherwise, it matches based on the name's string representation.
        If more than one body with the same name is found, an assertion error is raised.
        If no matching body is found, a `ValueError` is raised.

        :param name: The name of the DegreeOfFreedom to search for. Can be a string or a `PrefixedName` object.
        :return: The `DegreeOfFreedom` object that matches the given name.
        :raises ValueError: If multiple or no DegreeOfFreedom with the specified name are found.
        """
        if isinstance(name, PrefixedName):
            if name.prefix is not None:
                matches = [dof for dof in self.degrees_of_freedom if dof.name == name]
            else:
                matches = [
                    dof for dof in self.degrees_of_freedom if dof.name.name == name.name
                ]
        else:
            matches = [dof for dof in self.degrees_of_freedom if dof.name.name == name]
        if len(matches) > 1:
            raise ValueError(f"Multiple DegreeOfFreedom with name {name} found")
        if matches:
            return matches[0]
        raise KeyError(f"DegreeOfFreedom with name {name} not found")

    def get_connection_by_name(self, name: Union[str, PrefixedName]) -> Connection:
        """
        Retrieve a connection by its name.
        This method accepts either a string or a `PrefixedName` instance.
        It searches through the list of connections and returns the one
        that matches the given name. If the `PrefixedName` contains a prefix,
        the method ensures the name, including the prefix, matches an existing
        connection. Otherwise, it only considers the unprefixed name. If more than
        one connection matches the specified name, or if no connection is found,
        an exception is raised.

        :param name: The name of the connection to retrieve. Can be a string or
            a `PrefixedName` instance. If a prefix is included in `PrefixedName`,
            it will be used for matching.
        :return: The connection that matches the specified name.
        :raises ValueError: If multiple connections with the given name are found
            or if no connection with the given name exists.
        """
        if isinstance(name, PrefixedName):
            if name.prefix is not None:
                matches = [conn for conn in self.connections if conn.name == name]
            else:
                matches = [
                    conn for conn in self.connections if conn.name.name == name.name
                ]
        else:
            matches = [conn for conn in self.connections if conn.name.name == name]
        if len(matches) > 1:
            raise ValueError(f"Multiple connections with name {name} found")
        if matches:
            return matches[0]
        raise KeyError(f"Connection with name {name} not found")

    @lru_cache(maxsize=None)
    def compute_child_kinematic_structure_entities(
        self, kinematic_structure_entity: KinematicStructureEntity
    ) -> List[KinematicStructureEntity]:
        """
        Computes the child entities of a given KinematicStructureEntity in the world.
        :param kinematic_structure_entity: The KinematicStructureEntity for which to compute children.
        :return: A list of child KinematicStructureEntities.
        """
        return list(
            self.kinematic_structure.successors(kinematic_structure_entity.index)
        )

    @lru_cache(maxsize=None)
    def compute_parent_kinematic_structure_entity(
        self, kinematic_structure_entity: KinematicStructureEntity
    ) -> Optional[KinematicStructureEntity]:
        """
        Computes the parent KinematicStructureEntity of a given KinematicStructureEntity in the world.
        :param kinematic_structure_entity: The KinematicStructureEntity for which to compute the parent KinematicStructureEntity.
        :return: The parent KinematicStructureEntity of the given KinematicStructureEntity.
         If the given KinematicStructureEntity is the root, None is returned.
        """
        parent = self.kinematic_structure.predecessors(kinematic_structure_entity.index)
        if len(parent) == 0:
            return None
        return parent[0]

    @lru_cache(maxsize=None)
    def compute_parent_connection(
        self, kinematic_structure_entity: KinematicStructureEntity
    ) -> Optional[Connection]:
        """
        Computes the parent connection of a given KinematicStructureEntity in the world.
        :param kinematic_structure_entity: The entityKinematicStructureEntity for which to compute the parent connection.
        :return: The parent connection of the given KinematicStructureEntity.
        """
        parent = self.compute_parent_kinematic_structure_entity(
            kinematic_structure_entity
        )
        if parent is None:
            return None

        return self.kinematic_structure.get_edge_data(
            parent.index,
            kinematic_structure_entity.index,
        )

    @lru_cache(maxsize=None)
    def compute_chain_of_kinematic_structure_entities(
        self, root: KinematicStructureEntity, tip: KinematicStructureEntity
    ) -> List[KinematicStructureEntity]:
        """
        Computes the chain between root and tip. Can handle chains that start and end anywhere in the tree.
        """
        if root == tip:
            return [root]
        shortest_paths = rx.all_shortest_paths(
            self.kinematic_structure, root.index, tip.index, as_undirected=False
        )

        if len(shortest_paths) == 0:
            raise rx.NoPathFound(f"No path found from {root} to {tip}")

        return [self.kinematic_structure[index] for index in shortest_paths[0]]

    @lru_cache(maxsize=None)
    def compute_chain_of_connections(
        self, root: KinematicStructureEntity, tip: KinematicStructureEntity
    ) -> List[Connection]:
        """
        Computes the chain of connections between root and tip. Can handle chains that start and end anywhere in the tree.
        """
        entity_chain = self.compute_chain_of_kinematic_structure_entities(root, tip)
        return [
            self.get_connection(entity_chain[i], entity_chain[i + 1])
            for i in range(len(entity_chain) - 1)
        ]

    @lru_cache(maxsize=None)
    def compute_split_chain_of_kinematic_structure_entities(
        self, root: KinematicStructureEntity, tip: KinematicStructureEntity
    ) -> Tuple[
        List[KinematicStructureEntity],
        List[KinematicStructureEntity],
        List[KinematicStructureEntity],
    ]:
        """
        Computes the chain between root and tip. Can handle chains that start and end anywhere in the tree.
        :param root: The root KinematicStructureEntity to start the chain from
        :param tip: The tip KinematicStructureEntity to end the chain at
        :return: tuple containing
                    1. chain from root to the common ancestor (excluding common ancestor)
                    2. list containing just the common ancestor
                    3. chain from common ancestor to tip (excluding common ancestor)
        """
        if root == tip:
            return [], [root], []
        root_chain = self.compute_chain_of_kinematic_structure_entities(self.root, root)
        tip_chain = self.compute_chain_of_kinematic_structure_entities(self.root, tip)
        i = 0
        for i in range(min(len(root_chain), len(tip_chain))):
            if root_chain[i] != tip_chain[i]:
                break
        else:
            i += 1
        common_ancestor = tip_chain[i - 1]
        root_chain = self.compute_chain_of_kinematic_structure_entities(
            common_ancestor, root
        )
        root_chain = root_chain[1:]
        root_chain = root_chain[::-1]
        tip_chain = self.compute_chain_of_kinematic_structure_entities(
            common_ancestor, tip
        )
        tip_chain = tip_chain[1:]
        return root_chain, [common_ancestor], tip_chain

    @lru_cache(maxsize=None)
    def compute_split_chain_of_connections(
        self, root: KinematicStructureEntity, tip: KinematicStructureEntity
    ) -> Tuple[List[Connection], List[Connection]]:
        """
        Computes split chains of connections between 'root' and 'tip' bodies. Returns tuple of two Connection lists:
        (root->common ancestor, tip->common ancestor). Returns empty lists if root==tip.

        :param root: The starting `KinematicStructureEntity` object for the chain of connections.
        :param tip: The ending `KinematicStructureEntity` object for the chain of connections.
        :return: A tuple of two lists: the first list contains `Connection` objects from the `root` to
            the common ancestor, and the second list contains `Connection` objects from the `tip` to the
            common ancestor.
        """
        if root == tip:
            return [], []
        root_chain, common_ancestor, tip_chain = (
            self.compute_split_chain_of_kinematic_structure_entities(root, tip)
        )
        root_chain = root_chain + [common_ancestor[0]]
        tip_chain = [common_ancestor[0]] + tip_chain
        root_connections = []
        for i in range(len(root_chain) - 1):
            root_connections.append(
                self.get_connection(root_chain[i + 1], root_chain[i])
            )
        tip_connections = []
        for i in range(len(tip_chain) - 1):
            tip_connections.append(self.get_connection(tip_chain[i], tip_chain[i + 1]))
        return root_connections, tip_connections

    @property
    def kinematic_structure_entities_topologically_sorted(
        self,
    ) -> List[KinematicStructureEntity]:
        """
        Return a list of all kinematic_structure_entities in the world, sorted topologically.
        """
        indices = rx.topological_sort(self.kinematic_structure)

        return [self.kinematic_structure[index] for index in indices]

    @property
    def bodies_topologically_sorted(self) -> List[KinematicStructureEntity]:
        return [
            entity
            for entity in self.kinematic_structure_entities_topologically_sorted
            if isinstance(entity, KinematicStructureEntity)
        ]

    def copy_subgraph_to_new_world(self, new_root: KinematicStructureEntity) -> World:
        """
        Copies the subgraph of the kinematic structure from the root body to a new world and removes it from the old world.

        :param new_root: The root body of the subgraph to be copied.
        :return: A new `World` instance containing the copied subgraph.
        """
        with self.modify_world():
            new_world = World(name=self.name)
            child_bodies = self.compute_descendent_child_kinematic_structure_entities(
                new_root
            )
            child_body_parent_connections = [
                body.parent_connection for body in child_bodies
            ]

            with new_world.modify_world():
                self.remove_kinematic_structure_entity(new_root)
                new_world.add_kinematic_structure_entity(new_root)

                for body in child_bodies:
                    self.remove_kinematic_structure_entity(body)
                    new_world.add_kinematic_structure_entity(body)

                for connection in child_body_parent_connections:
                    self.remove_connection(connection)
                    new_world.add_connection(connection)

        return new_world

    @property
    def empty(self):
        """
        :return: Returns True if the world contains no kinematic_structure_entities, else False.
        """
        return len(self.kinematic_structure_entities) == 0

    @property
    def layers(self) -> List[List[KinematicStructureEntity]]:
        return rx.layers(
            self.kinematic_structure, [self.root.index], index_output=False
        )

    def bfs_layout(
        self, scale: float = 1.0, align: PlotAlignment = PlotAlignment.VERTICAL
    ) -> Dict[int, np.array]:
        """
        Generate a bfs layout for this circuit.

        :return: A dict mapping the node indices to 2d coordinates.
        """
        layers = self.layers

        pos = None
        nodes = []
        width = len(layers)
        for i, layer in enumerate(layers):
            height = len(layer)
            xs = np.repeat(i, height)
            ys = np.arange(0, height, dtype=float)
            offset = ((width - 1) / 2, (height - 1) / 2)
            layer_pos = np.column_stack([xs, ys]) - offset
            if pos is None:
                pos = layer_pos
            else:
                pos = np.concatenate([pos, layer_pos])
            nodes.extend(layer)

        # Find max length over all dimensions
        pos -= pos.mean(axis=0)
        lim = np.abs(pos).max()  # max coordinate for all axes
        # rescale to (-scale, scale) in all directions, preserves aspect
        if lim > 0:
            pos *= scale / lim

        if align == PlotAlignment.HORIZONTAL:
            pos = pos[:, ::-1]  # swap x and y coords

        pos = dict(zip([node.index for node in nodes], pos))
        return pos

    def plot_kinematic_structure(
        self, scale: float = 1.0, align: PlotAlignment = PlotAlignment.VERTICAL
    ) -> None:
        """
        Plots the kinematic structure of the world.
        The plot shows entities as nodes and connections as edges in a directed graph.
        """
        # Create a new figure
        plt.figure(figsize=(12, 8))

        pos = self.bfs_layout(scale=scale, align=align)

        rustworkx.visualization.mpl_draw(
            self.kinematic_structure,
            pos=pos,
            labels=lambda body: str(body.name),
            with_labels=True,
            edge_labels=lambda edge: edge.__class__.__name__,
        )

        plt.title("World Kinematic Structure")
        plt.axis("off")  # Hide axes
        plt.show()

    def _travel_branch(
        self,
        kinematic_structure_entity: KinematicStructureEntity,
        visitor: rustworkx.visit.DFSVisitor,
    ) -> None:
        """
        Apply a DFS Visitor to a subtree of the kinematic structure.

        :param kinematic_structure_entity: Starting point of the search
        :param visitor: This visitor to apply.
        """
        rx.dfs_search(
            self.kinematic_structure, [kinematic_structure_entity.index], visitor
        )

    def compile_forward_kinematics_expressions(self) -> None:
        """
        Traverse the kinematic structure and compile forward kinematics expressions for fast evaluation.
        """

        if self.empty:
            return

        new_fks = ForwardKinematicsVisitor(self)
        self._travel_branch(self.root, new_fks)
        new_fks.compile_forward_kinematics()
        self._fk_computer = new_fks

    def _recompute_forward_kinematics(self) -> None:
        self._fk_computer.recompute()

    @copy_lru_cache()
    def compose_forward_kinematics_expression(
        self, root: KinematicStructureEntity, tip: KinematicStructureEntity
    ) -> cas.TransformationMatrix:
        """
        :param root: The root KinematicStructureEntity in the kinematic chain.
            It determines the starting point of the forward kinematics calculation.
        :param tip: The tip KinematicStructureEntity in the kinematic chain.
            It determines the endpoint of the forward kinematics calculation.
        :return: An expression representing the computed forward kinematics of the tip KinematicStructureEntity relative to the root KinematicStructureEntity.
        """

        fk = cas.TransformationMatrix()
        root_chain, tip_chain = self.compute_split_chain_of_connections(root, tip)
        connection: Connection
        for connection in root_chain:
            tip_T_root = connection.origin_expression.inverse()
            fk = fk.dot(tip_T_root)
        for connection in tip_chain:
            fk = fk.dot(connection.origin_expression)
        fk.reference_frame = root
        fk.child_frame = tip
        return fk

    def compute_forward_kinematics(
        self, root: KinematicStructureEntity, tip: KinematicStructureEntity
    ) -> cas.TransformationMatrix:
        """
        Compute the forward kinematics from the root KinematicStructureEntity to the tip KinematicStructureEntity.

        Calculate the transformation matrix representing the pose of the
        tip KinematicStructureEntity relative to the root KinematicStructureEntity.

        :param root: Root KinematicStructureEntity for which the kinematics are computed.
        :param tip: Tip KinematicStructureEntity to which the kinematics are computed.
        :return: Transformation matrix representing the relative pose of the tip KinematicStructureEntity with respect to the root KinematicStructureEntity.
        """
        return cas.TransformationMatrix(
            data=self.compute_forward_kinematics_np(root, tip), reference_frame=root
        )

    def compute_forward_kinematics_np(
        self, root: KinematicStructureEntity, tip: KinematicStructureEntity
    ) -> NpMatrix4x4:
        """
        Compute the forward kinematics from the root KinematicStructureEntity to the tip KinematicStructureEntity, root_T_tip and return it as a 4x4 numpy ndarray.

        Calculate the transformation matrix representing the pose of the
        tip KinematicStructureEntity relative to the root KinematicStructureEntity, expressed as a numpy ndarray.

        :param root: Root KinematicStructureEntity for which the kinematics are computed.
        :param tip: Tip KinematicStructureEntity to which the kinematics are computed.
        :return: Transformation matrix representing the relative pose of the tip KinematicStructureEntity with respect to the root KinematicStructureEntity.
        """
        return self._fk_computer.compute_forward_kinematics_np(root, tip).copy()

    def compute_forward_kinematics_of_all_collision_bodies(self) -> np.ndarray:
        """
        Computes a 4 by X matrix, with the forward kinematics of all collision bodies stacked on top each other.
        The entries are sorted by name of body.
        """
        return self._fk_computer.collision_fks

    def transform(
        self,
        spatial_object: cas.GenericSpatialType,
        target_frame: KinematicStructureEntity,
    ) -> cas.GenericSpatialType:
        """
        Transform a given spatial object from its reference frame to a target frame.

        Calculate the transformation from the reference frame of the provided
        spatial object to the specified target frame. Apply the transformation
        differently depending on the type of the spatial object:

        - If the object is a Quaternion, compute its rotation matrix, transform it, and
          convert back to a Quaternion.
        - For other types, apply the transformation matrix directly.

        :param spatial_object: The spatial object to be transformed.
        :param target_frame: The target KinematicStructureEntity frame to which the spatial object should
            be transformed.
        :return: The spatial object transformed to the target frame. If the input object
            is a Quaternion, the returned object is a Quaternion. Otherwise, it is the
            transformed spatial object.
        """
        target_frame_T_reference_frame = self.compute_forward_kinematics(
            root=target_frame, tip=spatial_object.reference_frame
        )
        if isinstance(spatial_object, cas.Quaternion):
            reference_frame_R = spatial_object.to_rotation_matrix()
            target_frame_R = target_frame_T_reference_frame @ reference_frame_R
            return target_frame_R.to_quaternion()
        else:
            return target_frame_T_reference_frame @ spatial_object

    def compute_inverse_kinematics(
        self,
        root: KinematicStructureEntity,
        tip: KinematicStructureEntity,
        target: cas.TransformationMatrix,
        dt: float = 0.05,
        max_iterations: int = 200,
        translation_velocity: float = 0.2,
        rotation_velocity: float = 0.2,
    ) -> Dict[DegreeOfFreedom, float]:
        """
        Compute inverse kinematics using quadratic programming.

        :param root: Root KinematicStructureEntity of the kinematic chain
        :param tip: Tip KinematicStructureEntity of the kinematic chain
        :param target: Desired tip pose relative to the root KinematicStructureEntity
        :param dt: Time step for integration
        :param max_iterations: Maximum number of iterations
        :param translation_velocity: Maximum translation velocity
        :param rotation_velocity: Maximum rotation velocity
        :return: Dictionary mapping DOF names to their computed positions
        """
        ik_solver = InverseKinematicsSolver(self)
        return ik_solver.solve(
            root,
            tip,
            target,
            dt,
            max_iterations,
            translation_velocity,
            rotation_velocity,
        )

    def apply_control_commands(
        self, commands: np.ndarray, dt: float, derivative: Derivatives
    ) -> None:
        """
        Updates the state of a system by applying control commands at a specified derivative level,
        followed by backward integration to update lower derivatives.

        :param commands: Control commands to be applied at the specified derivative
            level. The array length must match the number of free variables
            in the system.
        :param dt: Time step used for the integration of lower derivatives.
        :param derivative: The derivative level to which the control commands are
            applied.
        :return: None
        """
        assert len(commands) == len(
            self.degrees_of_freedom
        ), f"Commands length {len(commands)} does not match number of free variables {len(self.degrees_of_freedom)}"

        self.state.set_derivative(derivative, commands)

        for i in range(derivative - 1, -1, -1):
            self.state.set_derivative(
                i,
                self.state.get_derivative(i) + self.state.get_derivative(i + 1) * dt,
            )
        for connection in self.connections:
            if isinstance(connection, HasUpdateState):
                connection.update_state(dt)
        self.notify_state_change()

    def set_positions_1DOF_connection(
        self, new_state: Dict[ActiveConnection1DOF, float]
    ) -> None:
        """
        Set the positions of 1DOF connections and notify the world of the state change.
        """
        for connection, value in new_state.items():
            connection.position = value
        self.notify_state_change()

    def __deepcopy__(self, memo):
        new_world = World(name=self.name)
        body_mapping = {}
        dof_mapping = {}
        with new_world.modify_world():
            for body in self.bodies:
                new_body = Body(
                    visual=body.visual,
                    collision=body.collision,
                    name=body.name,
                )
                new_world.add_kinematic_structure_entity(new_body)
                body_mapping[body] = new_body
            for dof in self.degrees_of_freedom:
                new_dof = DegreeOfFreedom(
                    name=dof.name,
                    lower_limits=dof.lower_limits,
                    upper_limits=dof.upper_limits,
                )
                new_world.add_degree_of_freedom(new_dof)
                dof_mapping[dof] = new_dof
            for connection in self.connections:
                new_connection = SubclassJSONSerializer.from_json(connection.to_json())
                new_connection.parent = (
                    new_world.get_kinematic_structure_entity_by_name(
                        new_connection.parent.name
                    )
                )
                new_connection.child = new_world.get_kinematic_structure_entity_by_name(
                    new_connection.child.name
                )
                new_world.add_connection(new_connection)
            for dof in self.degrees_of_freedom:
                new_world.state[dof.name] = self.state[dof.name].data
        return new_world

    def load_collision_srdf(self, file_path: str):
        """
        Creates a CollisionConfig instance from an SRDF file.

        Parse an SRDF file to configure disabled collision pairs or bodies for a given world.
        Process SRDF elements like `disable_collisions`, `disable_self_collision`,
        or `disable_all_collisions` to update collision configuration
        by referencing bodies in the provided `world`.

        :param file_path: The path to the SRDF file used for collision configuration.
        """
        SRDF_DISABLE_ALL_COLLISIONS: str = "disable_all_collisions"
        SRDF_DISABLE_SELF_COLLISION: str = "disable_self_collision"
        SRDF_MOVEIT_DISABLE_COLLISIONS: str = "disable_collisions"

        if not os.path.exists(file_path):
            raise ValueError(f"file {file_path} does not exist")
        srdf = etree.parse(file_path)
        srdf_root = srdf.getroot()
        for child in srdf_root:
            if hasattr(child, "tag"):
                if child.tag in {
                    SRDF_MOVEIT_DISABLE_COLLISIONS,
                    SRDF_DISABLE_SELF_COLLISION,
                }:
                    body_a_srdf_name: str = child.attrib["link1"]
                    body_b_srdf_name: str = child.attrib["link2"]
                    body_a: KinematicStructureEntity = (
                        self.get_kinematic_structure_entity_by_name(body_a_srdf_name)
                    )
                    body_b: KinematicStructureEntity = (
                        self.get_kinematic_structure_entity_by_name(body_b_srdf_name)
                    )
                    if not body_a.has_collision():
                        continue
                    if not body_b.has_collision():
                        continue
                    self.add_disabled_collision_pair(body_a, body_b)
                elif child.tag == SRDF_DISABLE_ALL_COLLISIONS:
                    body: KinematicStructureEntity = (
                        self.get_kinematic_structure_entity_by_name(
                            child.attrib["link"]
                        )
                    )
                    collision_config = CollisionCheckingConfig(disabled=True)
                    body.set_static_collision_config(collision_config)

    @property
    def controlled_connections(self) -> Set[ActiveConnection]:
        """
        A subset of the robot's connections that are controlled by a controller.
        """
        return set(
            c
            for c in self.connections
            if isinstance(c, ActiveConnection) and c.has_hardware_interface
        )

    def is_controlled_connection_in_chain(
        self, root: KinematicStructureEntity, tip: KinematicStructureEntity
    ) -> bool:
        root_part, tip_part = self.compute_split_chain_of_connections(root, tip)
        connections = root_part + tip_part
        for c in connections:
            if (
                isinstance(c, ActiveConnection)
                and c.has_hardware_interface
                and not c.frozen_for_collision_avoidance
            ):
                return True
        return False

    def disable_collisions_for_adjacent_bodies(self):
        """
        Computes pairs of bodies that should not be collision checked because they have no controlled connections
        between them.

        When all connections between two bodies are not controlled, these bodies cannot move relative to each
        other, so collision checking between them is unnecessary.

        :return: Set of body pairs that should have collisions disabled
        """
        body_combinations = set(
            combinations_with_replacement(self.bodies_with_enabled_collision, 2)
        )
        for body_a, body_b in list(body_combinations):
            if body_a == body_b:
                self.add_disabled_collision_pair(body_a, body_b)
                continue
            if self.is_controlled_connection_in_chain(body_a, body_b):
                continue
            self.add_disabled_collision_pair(body_a, body_b)

    @property
    def bodies_with_enabled_collision(self) -> List[Body]:
        return list(
            b
            for b in self.bodies
            if b.has_collision()
            and b.get_collision_config
            and not b.get_collision_config().disabled
        )

    @property
    def disabled_collision_pairs(
        self,
    ) -> Set[Tuple[KinematicStructureEntity, KinematicStructureEntity]]:
        return self._disabled_collision_pairs | self._temp_disabled_collision_pairs

    @property
    def enabled_collision_pairs(self) -> Set[Tuple[Body, Body]]:
        """
        The complement of disabled_collision_pairs with respect to all possible body combinations with enabled collision.
        """
        all_combinations = set(
            combinations_with_replacement(self.bodies_with_enabled_collision, 2)
        )
        return all_combinations - self.disabled_collision_pairs

    def add_disabled_collision_pair(
        self, body_a: KinematicStructureEntity, body_b: KinematicStructureEntity
    ):
        """
        Disable collision checking between two bodies
        """
        pair = tuple(sorted([body_a, body_b], key=lambda b: b.name))
        self._disabled_collision_pairs.add(pair)

    def add_temp_disabled_collision_pair(
        self, body_a: KinematicStructureEntity, body_b: KinematicStructureEntity
    ):
        """
        Disable collision checking between two bodies
        """
        pair = tuple(sorted([body_a, body_b], key=lambda b: b.name))
        self._temp_disabled_collision_pairs.add(pair)

    def get_direct_child_bodies_with_collision(
        self, connection: Connection
    ) -> Set[KinematicStructureEntity]:
        """
        Collect all child Bodies until a movable connection is found.

        :param connection: The connection from the kinematic structure whose child bodies will be traversed.
        :return: A set of Bodies that are moved directly by only this connection.
        """

        class BodyCollector(rx.visit.DFSVisitor):
            def __init__(self, world: World):
                self.world = world
                self.bodies = set()

            def discover_vertex(self, node_index: int, time: int) -> None:
                body = self.world.kinematic_structure[node_index]
                if body.has_collision():
                    self.bodies.add(body)

            def tree_edge(self, args: Tuple[int, int, Connection]) -> None:
                parent_index, child_index, e = args
                if (
                    isinstance(e, ActiveConnection)
                    and e.has_hardware_interface
                    and not e.frozen_for_collision_avoidance
                ):
                    raise rx.visit.PruneSearch()

        visitor = BodyCollector(self)
        rx.dfs_search(self.kinematic_structure, [connection.child.index], visitor)

        return visitor.bodies

    def compute_chain_reduced_to_controlled_joints(
        self, root: KinematicStructureEntity, tip: KinematicStructureEntity
    ) -> Tuple[KinematicStructureEntity, KinematicStructureEntity]:
        """
        Removes root and tip links until they are both connected with a controlled connection.
        Useful for implementing collision avoidance.

        1. Compute the kinematic chain of bodies between root and tip.
        2. Remove all entries from link_a downward until one is connected with a connection from this semantic annotation.
        2. Remove all entries from link_b upward until one is connected with a connection from this semantic annotation.

        :param root: start of the chain
        :param tip: end of the chain
        :return: start and end link of the reduced chain
        """
        downward_chain, upward_chain = self.compute_split_chain_of_connections(
            root=root, tip=tip
        )
        chain = downward_chain + upward_chain
        for i, connection in enumerate(chain):
            if (
                isinstance(connection, ActiveConnection)
                and connection.has_hardware_interface
                and not connection.frozen_for_collision_avoidance
            ):
                new_root = connection
                break
        else:
            raise KeyError(
                f"no controlled connection in chain between {root} and {tip}"
            )
        for i, connection in enumerate(reversed(chain)):
            if (
                isinstance(connection, ActiveConnection)
                and connection.has_hardware_interface
                and not connection.frozen_for_collision_avoidance
            ):
                new_tip = connection
                break
        else:
            raise KeyError(
                f"no controlled connection in chain between {root} and {tip}"
            )

        if new_root in upward_chain:
            new_root_body = new_root.parent
        else:  # if new_root is in the downward chain, we need to "flip" it by returning its child
            new_root_body = new_root.child
        if new_tip in upward_chain:
            new_tip_body = new_tip.child
        else:  # if new_root is in the downward chain, we need to "flip" it by returning its parent
            new_tip_body = new_tip.parent
        return new_root_body, new_tip_body

    def disable_non_robot_collisions(self) -> None:
        """
        Disables collision checks between bodies that do not belong to a robot.
        """
        robot_bodies = set()
        robot: AbstractRobot
        for robot in self.get_semantic_annotations_by_type(AbstractRobot):
            robot_bodies.update(robot.bodies_with_collisions)

        non_robot_bodies = set(self.bodies_with_enabled_collision) - robot_bodies
        for body_a in non_robot_bodies:
            for body_b in non_robot_bodies:
                self.add_disabled_collision_pair(body_a, body_b)

    def is_body_controlled(self, body: KinematicStructureEntity) -> bool:
        root_part, tip_part = self.compute_split_chain_of_connections(self.root, body)
        connections = root_part + tip_part
        for c in connections:
            if (
                isinstance(c, ActiveConnection)
                and c.has_hardware_interface
                and not c.frozen_for_collision_avoidance
            ):
                return True
        return False



from __future__ import annotations

import importlib
import inspect
import itertools
from abc import ABC
from collections import deque
from collections.abc import Iterable, Mapping
from dataclasses import dataclass, field
from dataclasses import fields
from functools import lru_cache

import numpy as np
import trimesh
import trimesh.boolean
from krrood.entity_query_language.predicate import Symbol
from random_events.utils import SubclassJSONSerializer
from scipy.stats import geom
from semantic_digital_twin.world_description.geometry import (
    transformation_to_json,
    transformation_from_json,
)
from trimesh.proximity import closest_point, nearby_faces
from trimesh.sample import sample_surface
from typing_extensions import (
    Deque,
    Type,
    TypeVar,
    Dict,
    Any,
    Self,
)
from typing_extensions import List, Optional, TYPE_CHECKING, Tuple
from typing_extensions import Set

from .geometry import TriangleMesh
from .shape_collection import ShapeCollection, BoundingBoxCollection
from ..datastructures.prefixed_name import PrefixedName
from ..exceptions import ReferenceFrameMismatchError
from ..spatial_types import spatial_types as cas
from ..spatial_types.spatial_types import TransformationMatrix, Expression, Point3
from ..utils import IDGenerator, type_string_to_type

if TYPE_CHECKING:
    from ..world_description.degree_of_freedom import DegreeOfFreedom
    from ..world import World

id_generator = IDGenerator()


@dataclass(unsafe_hash=True)
class WorldEntity(Symbol):
    """
    A class representing an entity in the world.
    """

    _world: Optional[World] = field(default=None, repr=False, kw_only=True, hash=False)
    """
    The backreference to the world this entity belongs to.
    """

    _semantic_annotations: Set[SemanticAnnotation] = field(
        default_factory=set, init=False, repr=False, hash=False
    )
    """
    The semantic annotations this entity is part of.
    """

    name: PrefixedName = field(default=None, kw_only=True)
    """
    The identifier for this world entity.
    """

    def __post_init__(self):
        if self.name is None:
            self.name = PrefixedName(f"{self.__class__.__name__}_{hash(self)}")


@dataclass
class CollisionCheckingConfig:
    buffer_zone_distance: Optional[float] = None
    """
    Distance defining a buffer zone around the entity. The buffer zone represents a soft boundary where
    proximity should be monitored but minor violations are acceptable.
    """

    violated_distance: float = 0.0
    """
    Critical distance threshold that must not be violated. Any proximity below this threshold represents
    a severe collision risk requiring immediate attention.
    """

    disabled: Optional[bool] = None
    """
    Flag to enable/disable collision checking for this entity. When True, all collision checks are ignored.
    """

    max_avoided_bodies: int = 1
    """
    Maximum number of other bodies this body should avoid simultaneously.
    If more bodies than this are in the buffer zone, only the closest ones are avoided.
    """


@dataclass(unsafe_hash=True)
class KinematicStructureEntity(WorldEntity, SubclassJSONSerializer, ABC):
    """
    An entity that is part of the kinematic structure of the world.
    """

    _world: Optional[World] = field(default=None, repr=False, hash=False, init=False)
    """
    Setting init=False because it should only be set by the World, not during initialization.
    """

    index: Optional[int] = field(default=None, init=False)
    """
    The index of the entity in `_world.kinematic_structure`.
    """

    @property
    def global_pose(self) -> TransformationMatrix:
        """
        Computes the pose of the KinematicStructureEntity in the world frame.
        :return: TransformationMatrix representing the global pose.
        """
        return self._world.compute_forward_kinematics(self._world.root, self)

    @property
    def parent_connection(self) -> Connection:
        """
        Returns the parent connection of this KinematicStructureEntity.
        """
        return self._world.compute_parent_connection(self)

    @property
    def child_kinematic_structure_entities(self) -> List[KinematicStructureEntity]:
        """
        Returns the direct child KinematicStructureEntity of this entity.
        """
        return self._world.compute_child_kinematic_structure_entities(self)

    @property
    def parent_kinematic_structure_entity(self) -> KinematicStructureEntity:
        """
        Returns the parent KinematicStructureEntity of this entity.
        """
        return self._world.compute_parent_kinematic_structure_entity(self)

    def get_first_parent_connection_of_type(
        self, connection_type: Type[GenericConnection]
    ) -> GenericConnection:
        """
        Traverse the chain up until an active connection is found.
        """
        if self == self._world.root:
            raise ValueError(
                f"Cannot get controlled parent connection for root body {self._world.root.name}."
            )
        if isinstance(self.parent_connection, connection_type):
            return self.parent_connection
        return self.parent_connection.parent.get_first_parent_connection_of_type(
            connection_type
        )


@dataclass
class Body(KinematicStructureEntity, SubclassJSONSerializer):
    """
    Represents a body in the world.
    A body is a semantic atom, meaning that it cannot be decomposed into meaningful smaller parts.
    """

    visual: ShapeCollection = field(default_factory=ShapeCollection, repr=False)
    """
    List of shapes that represent the visual appearance of the link.
    The poses of the shapes are relative to the link.
    """

    collision: ShapeCollection = field(default_factory=ShapeCollection, repr=False)
    """
    List of shapes that represent the collision geometry of the link.
    The poses of the shapes are relative to the link.
    """

    collision_config: Optional[CollisionCheckingConfig] = field(
        default_factory=CollisionCheckingConfig
    )
    """
    Configuration for collision checking.
    """

    temp_collision_config: Optional[CollisionCheckingConfig] = None
    """
    Temporary configuration for collision checking, takes priority over `collision_config`.
    """

    index: Optional[int] = field(default=None, init=False)
    """
    The index of the entity in `_world.kinematic_structure`.
    """

    def __post_init__(self):
        if not self.name:
            self.name = PrefixedName(f"body_{id_generator(self)}")

        self.visual.reference_frame = self
        self.collision.reference_frame = self
        self.collision.transform_all_shapes_to_own_frame()
        self.visual.transform_all_shapes_to_own_frame()

    def get_collision_config(self) -> CollisionCheckingConfig:
        if self.temp_collision_config is not None:
            return self.temp_collision_config
        return self.collision_config

    def set_static_collision_config(self, collision_config: CollisionCheckingConfig):
        merged_config = CollisionCheckingConfig(
            buffer_zone_distance=(
                collision_config.buffer_zone_distance
                if collision_config.buffer_zone_distance is not None
                else self.collision_config.buffer_zone_distance
            ),
            violated_distance=collision_config.violated_distance,
            disabled=(
                collision_config.disabled
                if collision_config.disabled is not None
                else self.collision_config.disabled
            ),
            max_avoided_bodies=collision_config.max_avoided_bodies,
        )
        self.collision_config = merged_config

    def set_static_collision_distances(
        self, buffer_zone_distance: float, violated_distance: float
    ):
        self.collision_config.buffer_zone_distance = buffer_zone_distance
        self.collision_config.violated_distance = violated_distance

    def set_temporary_collision_config(self, collision_config: CollisionCheckingConfig):
        self.temp_collision_config = collision_config

    def reset_temporary_collision_config(self):
        self.temp_collision_config = None

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        if other is None:
            return False
        return self.name == other.name and self._world is other._world

    def has_collision(
        self, volume_threshold: float = 1.001e-6, surface_threshold: float = 0.00061
    ) -> bool:
        """
        Check if collision geometry is mesh or simple shape with volume/surface bigger than thresholds.

        :param volume_threshold: Ignore simple geometry shapes with a volume less than this (in m^3)
        :param surface_threshold: Ignore simple geometry shapes with a surface area less than this (in m^2)
        :return: True if collision geometry is mesh or simple shape exceeding thresholds
        """
        return len(self.collision) > 0

    def compute_closest_points_multi(
        self, others: list[Body], sample_size=25
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Computes the closest points to each given body respectively.

        :param others: The list of bodies to compute the closest points to.
        :param sample_size: The number of samples to take from the surface of the other bodies.
        :return: A tuple containing: The points on the self body, the points on the other bodies, and the distances. All points are in the of this body.
        """

        @lru_cache(maxsize=None)
        def evaluated_geometric_distribution(n: int) -> np.ndarray:
            """
            Evaluates the geometric distribution for a given number of samples.
            :param n: The number of samples to evaluate.
            :return: An array of probabilities for each sample.
            """
            return geom.pmf(np.arange(1, n + 1), 0.5)

        query_points = []
        for other in others:
            # Calculate the closest vertex on this body to the other body
            closest_vert_id = self.collision[0].mesh.kdtree.query(
                (
                    self._world.compute_forward_kinematics_np(self, other)
                    @ other.collision[0].origin.to_np()
                )[:3, 3],
                k=1,
            )[1]
            closest_vert = self.collision[0].mesh.vertices[closest_vert_id]

            # Compute the closest faces on the other body to the closes vertex
            faces = nearby_faces(
                other.collision[0].mesh,
                [
                    (
                        self._world.compute_forward_kinematics_np(other, self)
                        @ self.collision[0].origin.to_np()
                    )[:3, 3]
                    + closest_vert
                ],
            )[0]
            face_weights = np.zeros(len(other.collision[0].mesh.faces))

            # Assign weights to the faces based on a geometric distribution
            face_weights[faces] = evaluated_geometric_distribution(len(faces))

            # Sample points on the surface of the other body
            q = sample_surface(
                other.collision[0].mesh, sample_size, face_weight=face_weights, seed=420
            )[0]
            # Make 4x4 transformation matrix from points
            points = np.tile(np.eye(4, dtype=np.float32), (len(q), 1, 1))
            points[:, :3, 3] = q

            # Transform from the mesh to the other mesh
            transform = (
                np.linalg.inv(self.collision[0].origin.to_np())
                @ self._world.compute_forward_kinematics_np(self, other)
                @ other.collision[0].origin.to_np()
            )
            points = points @ transform

            points = points[
                :, :3, 3
            ]  # Extract the points from the transformation matrix

            query_points.extend(points)

        # Actually compute the closest points
        points, dists = closest_point(self.collision[0].mesh, query_points)[:2]
        # Find the closest points for each body out of all the sampled points
        points = np.array(points).reshape(len(others), sample_size, 3)
        dists = np.array(dists).reshape(len(others), sample_size)
        dist_min = np.min(dists, axis=1)
        points_min_self = points[np.arange(len(others)), np.argmin(dists, axis=1), :]
        points_min_other = np.array(query_points).reshape(len(others), sample_size, 3)[
            np.arange(len(others)), np.argmin(dists, axis=1), :
        ]
        return points_min_self, points_min_other, dist_min

    def to_json(self) -> Dict[str, Any]:
        result = super().to_json()
        result["name"] = self.name.to_json()
        result["collision"] = self.collision.to_json()
        result["visual"] = self.visual.to_json()
        return result

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self:

        result = cls(name=PrefixedName.from_json(data["name"]))

        collision = ShapeCollection.from_json(data["collision"])
        visual = ShapeCollection.from_json(data["visual"])

        for shape in itertools.chain(collision, visual):
            shape.origin.reference_frame = result

        result.collision = collision
        result.visual = visual

        return result


@dataclass
class Region(KinematicStructureEntity):
    """
    Virtual KinematicStructureEntity representing a semantic region in the world.
    """

    area: ShapeCollection = field(default_factory=ShapeCollection, hash=False)
    """
    The shapes that represent the area of the region.
    """

    def __post_init__(self):
        self.area.reference_frame = self

    def __hash__(self):
        return id(self)

    @classmethod
    def from_3d_points(
        cls,
        name: PrefixedName,
        points_3d: List[Point3],
        reference_frame: Optional[Body] = None,
        minimum_thickness: float = 0.005,
        sv_ratio_tol: float = 1e-7,
    ) -> Self:
        """
        Constructs a Region from a list of 3D points by creating a convex hull around them.
        The points are analyzed to determine if they are approximately planar. If they are,
        a minimum thickness is added to ensure the region has a non-zero volume.

        :param name: Prefixed name for the region.
        :param points_3d: List of 3D points.
        :param reference_frame: Optional reference frame.
        :param minimum_thickness: Minimum thickness to add if points are near-planar.
        :param sv_ratio_tol: Tolerance for determining planarity based on singular value ratio.

        :return: Region object.
        """
        points = np.asarray([point.to_np()[:3] for point in points_3d], dtype=float)
        points = np.unique(points, axis=0)
        assert (
            len(points) >= 3
        ), "At least 4 unique points are required to define a 3D region."

        centered_points = points - points.mean(axis=0, keepdims=True)
        assert np.any(centered_points), "Points must not be all identical."

        # We compute the principal axes of the point cloud using SVD.
        # This allows us to reason about the geometric thickness of our point cloud.
        # The axis with the smallest variance, located at the last index if our `principal_axis` is our `normal`
        # indicating the direction of the region's thickness.
        _, variance, principal_axis = np.linalg.svd(
            centered_points, full_matrices=False
        )
        smallest_variance_axis = principal_axis[-1]  # this is our normal
        unit_vector_normal = smallest_variance_axis / np.linalg.norm(
            smallest_variance_axis
        )

        # We compute the thickness, peak-to-peak (max - min), along the normal direction, to get the thickness of
        # the region.
        thickness_in_normal_direction = np.ptp(centered_points @ unit_vector_normal)
        is_near_planar = variance[0] > 0 and variance[-1] / variance[0] < sv_ratio_tol
        thickness_padding = (
            minimum_thickness / 2
            if thickness_in_normal_direction < minimum_thickness or is_near_planar
            else 0.0
        )

        # We do not provide any 2d shapes, since they would be very weird to handle with raytracing etc.
        # Thus we decided that in near-planar cases we add a minimum thickness to ensure we get a 3d shape.
        if thickness_padding > 0:
            P_aug = np.vstack(
                [
                    points + thickness_padding * unit_vector_normal,
                    points - thickness_padding * unit_vector_normal,
                ]
            )
        else:
            P_aug = points

        hull = trimesh.points.PointCloud(P_aug).convex_hull
        hull.remove_unreferenced_vertices()
        hull.remove_degenerate_faces()
        hull.process()

        area_mesh = TriangleMesh(
            mesh=hull, origin=TransformationMatrix(reference_frame=reference_frame)
        )
        return cls(name=name, area=ShapeCollection([area_mesh]))

    def to_json(self) -> Dict[str, Any]:
        result = super().to_json()
        result["name"] = self.name.to_json()
        result["area"] = self.area.to_json()
        return result

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self:
        result = cls(name=PrefixedName.from_json(data["name"]))
        area = ShapeCollection.from_json(data["area"])
        for shape in area:
            shape.origin.reference_frame = result
        result.area = area
        return result


GenericKinematicStructureEntity = TypeVar(
    "GenericKinematicStructureEntity", bound=KinematicStructureEntity
)


@dataclass
class SemanticAnnotation(WorldEntity, SubclassJSONSerializer):
    """
    Represents a semantic annotation on a set of bodies in the world.

    This class can hold references to certain bodies that gain meaning in this context.

    .. warning::

        The hash of a semantic annotation is based on the hash of its type and kinematic structure entities.
        Overwrite this with extreme care and only if you know what you are doing. Hashes are used inside rules to check if
        a new semantic annotation has been created. If you, for instance, just use the object identity, this will fail since python assigns
        new memory pointers always. The same holds for the equality operator.
        If you do not want to change the behavior, make sure to use @dataclass(eq=False) to decorate your class.
    """

    def __post_init__(self):
        if self.name is None:
            self.name = PrefixedName(
                name=f"{self.__class__.__name__}_{id_generator(self)}",
                prefix=self._world.name if self._world is not None else None,
            )

    def __hash__(self):
        return hash(
            tuple(
                [self.__class__]
                + sorted([kse.name for kse in self.kinematic_structure_entities])
            )
        )

    def __eq__(self, other):
        return hash(self) == hash(other)

    def to_json(self) -> Dict[str, Any]:
        result = {
            **super().to_json(),
        }

        for semantic_annotation_field in fields(self):
            value = getattr(self, semantic_annotation_field.name)
            if issubclass(type(value), SubclassJSONSerializer):
                result[semantic_annotation_field.name] = value.to_json()
        return result

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self:
        semantic_annotation_fields = {f.name: f for f in fields(cls)}

        init_args = {}

        for k, v in semantic_annotation_fields.items():
            if k not in data.keys():
                continue
            field_type = type_string_to_type(data[k]["type"])
            if issubclass(field_type, SubclassJSONSerializer):
                init_args[k] = field_type.from_json(data[k])

        return cls(**init_args)

    def _kinematic_structure_entities(
        self, visited: Set[int], aggregation_type: Type[GenericKinematicStructureEntity]
    ) -> Set[GenericKinematicStructureEntity]:
        """
        Recursively collects all entities that are part of this semantic annotation.
        """
        stack: Deque[object] = deque([self])
        entities: Set[aggregation_type] = set()

        while stack:
            obj = stack.pop()
            oid = id(obj)
            if oid in visited:
                continue
            visited.add(oid)

            match obj:
                case aggregation_type():
                    entities.add(obj)

                case SemanticAnnotation():
                    stack.extend(_attr_values(obj, aggregation_type))

                case Mapping():
                    stack.extend(
                        v
                        for v in obj.values()
                        if _is_entity_semantic_annotation_or_iterable(
                            v, aggregation_type
                        )
                    )

                case Iterable() if not isinstance(obj, (str, bytes, bytearray)):
                    stack.extend(
                        v
                        for v in obj
                        if _is_entity_semantic_annotation_or_iterable(
                            v, aggregation_type
                        )
                    )

        return entities

    @property
    def kinematic_structure_entities(self) -> Iterable[KinematicStructureEntity]:
        """
        Returns a Iterable of all relevant KinematicStructureEntity in this semantic annotation. The default behaviour is to aggregate all KinematicStructureEntity that are accessible
        through the properties and fields of this semantic annotation, recursively.
        If this behaviour is not desired for a specific semantic annotation, it can be overridden by implementing the `KinematicStructureEntity` property.
        """
        return self._kinematic_structure_entities(set(), KinematicStructureEntity)

    @property
    def bodies(self) -> Iterable[Body]:
        """
        Returns an Iterable of all relevant bodies in this semantic annotation. The default behaviour is to aggregate all bodies that are accessible
        through the properties and fields of this semantic annotation, recursively.
        If this behaviour is not desired for a specific semantic annotation, it can be overridden by implementing the `bodies` property.
        """
        return self._kinematic_structure_entities(set(), Body)

    @property
    def regions(self) -> Iterable[Region]:
        """
        Returns an Iterable of all relevant regions in this semantic annotation. The default behaviour is to aggregate all regions that are accessible
        through the properties and fields of this semantic annotation, recursively.
        If this behaviour is not desired for a specific semantic annotation, it can be overridden by implementing the `regions` property.
        """
        return self._kinematic_structure_entities(set(), Region)

    def as_bounding_box_collection_at_origin(
        self, origin: TransformationMatrix
    ) -> BoundingBoxCollection:
        """
        Returns a bounding box collection that contains the bounding boxes of all bodies in this semantic annotation.
        :param reference_frame: The reference frame to express the bounding boxes in.
        :returns: A collection of bounding boxes in world-space coordinates.
        """

        collections = iter(
            entity.collision.as_bounding_box_collection_at_origin(origin)
            for entity in self.kinematic_structure_entities
            if isinstance(entity, Body) and entity.has_collision()
        )
        bbs = BoundingBoxCollection([], origin.reference_frame)

        for bb_collection in collections:
            bbs = bbs.merge(bb_collection)

        return bbs

    def as_bounding_box_collection_in_frame(
        self, reference_frame: KinematicStructureEntity
    ) -> BoundingBoxCollection:
        """
        Provides the bounding box collection for this entity in the given reference frame.
        :param reference_frame: The reference frame to express the bounding boxes in.
        :returns: A collection of bounding boxes in world-space coordinates.
        """
        return self.as_bounding_box_collection_at_origin(
            TransformationMatrix(reference_frame=reference_frame)
        )


@dataclass(eq=False)
class RootedSemanticAnnotation(SemanticAnnotation):
    """
    Represents a semantic annotation that is rooted in a specific KinematicStructureEntity.
    """

    root: Body = field(default=None)

    @property
    def connections(self) -> List[Connection]:
        return self._world.get_connections_of_branch(self.root)

    @property
    def bodies(self) -> List[Body]:
        return self._world.get_kinematic_structure_entities_of_branch(self.root)

    @property
    def bodies_with_collisions(self) -> List[Body]:
        return [x for x in self.bodies if x.has_collision()]

    @property
    def bodies_with_enabled_collision(self) -> Set[Body]:
        return set(
            body
            for body in self.bodies
            if body.has_collision() and not body.get_collision_config().disabled
        )


@dataclass(eq=False)
class SemanticEnvironmentAnnotation(RootedSemanticAnnotation):
    """
    Represents a semantic annotation of the environment.
    """

    @property
    def kinematic_structure_entities(self) -> Set[KinematicStructureEntity]:
        """
        Returns a set of all KinematicStructureEntity in the environment semantic annotation.
        """
        return set(
            self._world.get_kinematic_structure_entities_of_branch(self.root)
        ) | {self.root}


@dataclass
class Connection(WorldEntity, SubclassJSONSerializer):
    """
    Represents a connection between two entities in the world.
    """

    _world: Optional[World] = field(default=None, repr=False, hash=False, init=False)
    """
    Setting init=False because it should only be set by the World, not during initialization.
    """

    parent: KinematicStructureEntity
    """
    The parent KinematicStructureEntity of the connection.
    """

    child: KinematicStructureEntity
    """
    The child KinematicStructureEntity of the connection.
    """

    parent_T_connection_expression: TransformationMatrix = field(default=None)
    connection_T_child_expression: TransformationMatrix = field(default=None)
    """
    The origin expression of a connection is split into 2 transforms:
    1. parent_T_connection describes the pose of the connection and is always constant.
       It typically describes the fixed part of the origin expression, equivalent to the origin tag in urdf. 
       For example, it is the point about which a revolute joint rotates.
    2. connection_T_child describes the pose of the child relative to the connection.
       This typically contains only the expressions that describe how the degrees of freedom move the child.
       For example, it describes how the angle of a revolute joint affects the child pose.

    This split is necessary for copying Connections, because they need parent_T_connection as an input parameter and 
    connection_T_child is generated in the __post_init__ method.
    """

    def to_json(self) -> Dict[str, Any]:
        result = super().to_json()
        result["name"] = self.name.to_json()
        result["parent"] = self.parent.to_json()
        result["child"] = self.child.to_json()
        result["parent_T_connection_expression"] = transformation_to_json(
            self.parent_T_connection_expression
        )
        return result

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self:
        return cls(
            name=PrefixedName.from_json(data["name"]),
            parent=KinematicStructureEntity.from_json(data["parent"]),
            child=KinematicStructureEntity.from_json(data["child"]),
            parent_T_connection_expression=transformation_from_json(
                data["parent_T_connection_expression"]
            ),
        )

    @property
    def origin_expression(self) -> TransformationMatrix:
        return self.parent_T_connection_expression @ self.connection_T_child_expression

    @property
    def has_hardware_interface(self) -> bool:
        return False

    def add_to_world(self, world: World):
        self._world = world

    def __post_init__(self):
        self.name = self.name or self._generate_default_name(
            parent=self.parent, child=self.child
        )

        # If I use default factories, I'd have to complicate the from_json, because I couldn't blindly pass these args
        if self.parent_T_connection_expression is None:
            self.parent_T_connection_expression = TransformationMatrix()
        if self.connection_T_child_expression is None:
            self.connection_T_child_expression = TransformationMatrix()

        if (
            self.parent_T_connection_expression.reference_frame is not None
            and self.parent_T_connection_expression.reference_frame != self.parent
        ):
            raise ReferenceFrameMismatchError(
                self.parent, self.parent_T_connection_expression.reference_frame
            )

        self.parent_T_connection_expression.reference_frame = self.parent
        self.connection_T_child_expression.child_frame = self.child

    @classmethod
    def _generate_default_name(
        cls, parent: KinematicStructureEntity, child: KinematicStructureEntity
    ) -> PrefixedName:
        return PrefixedName(
            f"{parent.name.name}_T_{child.name.name}", prefix=child.name.prefix
        )

    def __hash__(self):
        return hash((self.parent, self.child))

    def __eq__(self, other):
        return self.name == other.name

    @property
    def origin(self) -> cas.TransformationMatrix:
        """
        :return: The relative transform between the parent and child frame.
        """
        return self._world.compute_forward_kinematics(self.parent, self.child)

    def origin_as_position_quaternion(self) -> Expression:
        position = self.origin_expression.to_position()[:3]
        orientation = self.origin_expression.to_quaternion()
        return cas.Expression.vstack([position, orientation]).T

    @property
    def dofs(self) -> Set[DegreeOfFreedom]:
        """
        Returns the degrees of freedom associated with this connection.
        """
        dofs = set()

        if hasattr(self, "active_dofs"):
            dofs.update(set(self.active_dofs))
        if hasattr(self, "passive_dofs"):
            dofs.update(set(self.passive_dofs))

        return dofs

    @classmethod
    def create_with_dofs(
        cls,
        world: World,
        parent: KinematicStructureEntity,
        child: KinematicStructureEntity,
        name: Optional[PrefixedName] = None,
        *args,
        **kwargs,
    ) -> Self:
        """
        This method will automatically generate the degrees of freedom for this connection into the given world
        and initialize the connection with the generated dofs.
        :param world: Reference to the world where the dofs should be added.
        :param parent: parent of the connection.
        :param child: child of the connection.
        :param name: name of the connection.
        :param args: additional arguments the subclass might need
        :param kwargs: additional keyword arguments the subclass might need
        :return:
        """
        raise NotImplementedError(
            "ConnectionWithDofs.create_with_dofs is not implemented."
        )


GenericConnection = TypeVar("GenericConnection", bound=Connection)


def _is_entity_semantic_annotation_or_iterable(
    obj: object, aggregation_type: Type[KinematicStructureEntity]
) -> bool:
    """
    Determines if an object is a KinematicStructureEntity, a semantic annotation, or an Iterable (excluding strings and bytes).
    """
    return isinstance(obj, (aggregation_type, SemanticAnnotation)) or (
        isinstance(obj, Iterable) and not isinstance(obj, (str, bytes, bytearray))
    )


def _attr_values(
    semantic_annotation: SemanticAnnotation,
    aggregation_type: Type[GenericKinematicStructureEntity],
) -> Iterable[object]:
    """
    Yields all dataclass fields and set properties of this semantic annotation.
    Skips private fields (those starting with '_'), as well as the 'bodies' property.

    :param semantic_annotation: The semantic annotation to extract attributes from.
    """
    for f in fields(semantic_annotation):
        if f.name.startswith("_"):
            continue
        v = getattr(semantic_annotation, f.name, None)
        if _is_entity_semantic_annotation_or_iterable(v, aggregation_type):
            yield v

    for name, prop in inspect.getmembers(
        type(semantic_annotation), lambda o: isinstance(o, property)
    ):
        if name in {
            "kinematic_structure_entities",
            "bodies",
            "regions",
        } or name.startswith("_"):
            continue
        try:
            v = getattr(semantic_annotation, name)
        except Exception:
            continue
        if _is_entity_semantic_annotation_or_iterable(v, aggregation_type):
            yield v



import logging
import re
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import List, Callable

import numpy as np

from ..spatial_types import Point3
from ..spatial_types.spatial_types import TransformationMatrix
from ..semantic_annotations.factories import SemanticAnnotationFactory
from ..world import World
from ..world_description.geometry import TriangleMesh, FileMesh
from ..world_description.world_entity import Body


@dataclass
class Step(ABC):
    """
    A Step is a transformation that takes a World as input and produces a modified World as output.
    Steps are intended to be used in a Pipeline, where the output World of one Step is passed as the input World to the next Step.
    """

    @abstractmethod
    def _apply(self, world: World) -> World: ...

    def apply(self, world: World) -> World:
        with world.modify_world():
            return self._apply(world)


@dataclass
class Pipeline:
    """
    A Pipeline is a sequence of Steps that are applied to a World in order.
    Each Step takes the World as input and produces a modified World as output.
    The output World of one Step is passed as the input World to the next Step.
    """

    steps: List[Step]
    """
    The list of Steps to be applied in the Pipeline.
    """

    def apply(self, world: World) -> World:
        for step in self.steps:
            world = step.apply(world)
        return world


@dataclass
class BodyFilter(Step):
    """
    Filters bodies in the world based on a given condition.
    """

    condition: Callable[[Body], bool]

    def _apply(self, world: World) -> World:
        for body in world.bodies:
            if not self.condition(body):
                world.remove_kinematic_structure_entity(body)
        return world


@dataclass
class CenterLocalGeometryAndPreserveWorldPose(Step):
    """
    Adjusts the vertices of the collision meshes of each body in the world so that the origin is at the center of the
    mesh, and then updates the parent connection of the body to preserve the original world pose.
    An example where this is useful is when parsing FBX files where all bodies in the resulting world have an origin
    at (0, 0, 0), even through the collision meshes are not centered around that point.
    """

    def _apply(self, world: World) -> World:
        for body in world.bodies_topologically_sorted:

            vertices = []

            for coll in body.collision:
                if isinstance(coll, (FileMesh, TriangleMesh)):
                    mesh = coll.mesh
                    if mesh.vertices.shape[0] > 0:
                        vertices.append(mesh.vertices.copy())

            if len(vertices) == 0:
                logging.warning(
                    f"Body {body.name.name} has no vertices in visual or collision shapes, skipping."
                )
                continue

            # Compute the axis-aligned bounding box center of all vertices
            all_vertices = np.vstack(vertices)
            mins = all_vertices.min(axis=0)
            maxs = all_vertices.max(axis=0)
            center = (mins + maxs) / 2.0

            for coll in body.collision:
                if isinstance(coll, (FileMesh, TriangleMesh)):
                    m = coll.mesh
                    if m.vertices.shape[0] > 0:
                        m.vertices -= center

            old_origin_T_new_origin = TransformationMatrix.from_point_rotation_matrix(
                Point3(*center)
            )

            if body.parent_connection:
                parent_T_old_origin = (
                    body.parent_connection.parent_T_connection_expression
                )

                body.parent_connection.parent_T_connection_expression = (
                    parent_T_old_origin @ old_origin_T_new_origin
                )

            for child in world.compute_child_kinematic_structure_entities(body):
                old_origin_T_child_origin = (
                    child.parent_connection.parent_T_connection_expression
                )
                child.parent_connection.parent_T_connection_expression = (
                    old_origin_T_new_origin.inverse() @ old_origin_T_child_origin
                )
        return world


@dataclass
class BodyFactoryReplace(Step):
    """
    Replace bodies in the world that match a given condition with new structures created by a factory.
    """

    body_condition: Callable[[Body], bool] = lambda x: bool(
        re.compile(r"^dresser_\d+.*$").fullmatch(x.name.name)
    )
    """
    Condition to filter bodies that should be replaced. Defaults to matching bodies containing "dresser_" followed by digits in their name.
    """

    factory_creator: Callable[[Body], SemanticAnnotationFactory] = None
    """
    A callable that takes a Body and returns a ViewFactory to create the new structure.
    """

    def _apply(self, world: World) -> World:
        filtered_bodies = [body for body in world.bodies if self.body_condition(body)]

        for body in filtered_bodies:
            factory = self.factory_creator(body)
            parent_connection = body.parent_connection
            if parent_connection is None:
                return factory.create()

            for entity in world.get_kinematic_structure_entities_of_branch(body):
                world.remove_kinematic_structure_entity(entity)

            world.remove_kinematic_structure_entity(body)

            new_world = factory.create()
            parent_connection.child = new_world.root
            world.merge_world(new_world, parent_connection)

        return world



import keyword
from typing import Union, Iterable

import hypothesis.strategies as st
import numpy as np
from hypothesis import assume
from hypothesis.strategies import composite
from numpy import pi

import semantic_digital_twin.spatial_types.spatial_types as cas
from .reference_implementations import shortest_angular_distance

BIG_NUMBER = 1e100
SMALL_NUMBER = 1e-100

all_expressions_float_np = Union[
    cas.SymbolicType, float, np.ndarray, Iterable[float], Iterable[Iterable[float]]
]


def to_float_or_np(x: all_expressions_float_np) -> Union[float, np.ndarray]:
    if isinstance(x, cas.SymbolicType):
        return x.to_np()
    return x


def assert_allclose(
    a: all_expressions_float_np,
    b: all_expressions_float_np,
    atol: float = 1e-3,
    rtol: float = 1e-3,
    equal_nan: bool = False,
):
    a = to_float_or_np(a)
    b = to_float_or_np(b)
    assert np.allclose(a, b, atol=atol, rtol=rtol, equal_nan=equal_nan)


def vector(x):
    return st.lists(float_no_nan_no_inf(), min_size=x, max_size=x)


def angle_positive():
    return st.floats(0, 2 * np.pi)


def random_angle():
    return st.floats(-np.pi, np.pi)


def compare_axis_angle(
    actual_angle: all_expressions_float_np,
    actual_axis: all_expressions_float_np,
    expected_angle: cas.ScalarData,
    expected_axis: all_expressions_float_np,
):
    actual_angle = to_float_or_np(actual_angle)
    actual_axis = to_float_or_np(actual_axis)
    expected_angle = to_float_or_np(expected_angle)
    expected_axis = to_float_or_np(expected_axis)
    try:
        assert_allclose(actual_axis, expected_axis)
        assert_allclose(shortest_angular_distance(actual_angle, expected_angle), 0.0)
    except AssertionError:
        try:
            assert_allclose(actual_axis, -expected_axis)
            assert_allclose(
                shortest_angular_distance(actual_angle, abs(expected_angle - 2 * pi)),
                0.0,
            )
        except AssertionError:
            assert_allclose(shortest_angular_distance(actual_angle, 0), 0.0)
            assert_allclose(shortest_angular_distance(0, expected_angle), 0.0)
            assert not np.any(np.isnan(actual_axis))
            assert not np.any(np.isnan(expected_axis))


def compare_orientations(
    actual_orientation: all_expressions_float_np,
    desired_orientation: all_expressions_float_np,
) -> None:
    try:
        assert_allclose(actual_orientation, desired_orientation)
    except:
        assert_allclose(actual_orientation, -desired_orientation)


@composite
def variable_name(draw):
    variable = draw(st.text("qwertyuiopasdfghjklzxcvbnm", min_size=1))
    assume(variable not in keyword.kwlist)
    return variable


@composite
def lists_of_same_length(
    draw, data_types=(), min_length=1, max_length=10, unique=False
):
    length = draw(st.integers(min_value=min_length, max_value=max_length))
    lists = []
    for elements in data_types:
        lists.append(
            draw(st.lists(elements, min_size=length, max_size=length, unique=unique))
        )
    return lists


@composite
def rnd_joint_state(draw, joint_limits):
    return {
        jn: draw(st.floats(ll, ul, allow_nan=False, allow_infinity=False))
        for jn, (ll, ul) in joint_limits.items()
    }


@composite
def rnd_joint_state2(draw, joint_limits):
    muh = draw(joint_limits)
    muh = {
        jn: ((ll if ll is not None else pi * 2), (ul if ul is not None else pi * 2))
        for (jn, (ll, ul)) in muh.items()
    }
    return {
        jn: draw(st.floats(ll, ul, allow_nan=False, allow_infinity=False))
        for jn, (ll, ul) in muh.items()
    }


def float_no_nan_no_inf(outer_limit=1e5):
    return float_no_nan_no_inf_min_max(-outer_limit, outer_limit)


def float_no_nan_no_inf_min_max(min_value=-1e5, max_value=1e5):
    return st.floats(
        allow_nan=False,
        allow_infinity=False,
        max_value=max_value,
        min_value=min_value,
        allow_subnormal=False,
    )


@composite
def sq_matrix(draw):
    i = draw(st.integers(min_value=1, max_value=10))
    i_sq = i**2
    l = draw(
        st.lists(float_no_nan_no_inf(outer_limit=1000), min_size=i_sq, max_size=i_sq)
    )
    return np.array(l).reshape((i, i))


def unit_vector(length, elements=None):
    if elements is None:
        elements = float_no_nan_no_inf()
    vector = st.lists(elements, min_size=length, max_size=length).filter(
        lambda x: SMALL_NUMBER < np.linalg.norm(x) < BIG_NUMBER
    )

    def normalize(v):
        v = [round(x, 4) for x in v]
        l = np.linalg.norm(v)
        if l == 0:
            return np.array([0] * (length - 1) + [1])
        return np.array([x / l for x in v])

    return st.builds(normalize, vector)


def quaternion():
    return unit_vector(4, float_no_nan_no_inf(outer_limit=1))



import json
import logging
import math
from dataclasses import dataclass, field
from functools import cached_property
from typing import Dict, Tuple, Union, Set, Optional, List

import numpy as np
from krrood.entity_query_language.entity import the, entity, let
from krrood.ormatic.eql_interface import eql_to_sql
from sqlalchemy.exc import NoResultFound
from sqlalchemy.orm import Session

from ...datastructures.prefixed_name import PrefixedName
from ...orm.model import WorldMapping
from ...orm.ormatic_interface import *
from ...spatial_types.spatial_types import (
    TransformationMatrix,
    Point3,
)
from ...semantic_annotations.factories import (
    DoorFactory,
    RoomFactory,
    WallFactory,
    HandleFactory,
    Direction,
    DoubleDoorFactory,
    HorizontalSemanticDirection,
    SemanticPositionDescription,
    VerticalSemanticDirection,
)
from ...world import World
from ...world_description.connections import FixedConnection
from ...world_description.geometry import Scale
from ...world_description.world_entity import Body


@dataclass
class ProcthorDoor:
    """
    Processes a door dictionary from Procthor, extracting the door's hole polygon and computing its scale and
    transformation matrix relative to the parent wall's horizontal center.
    """

    door_dict: dict
    """
    Dictionary representing a door from Procthors' JSON format
    """

    parent_wall_width: float
    """
    Width of the parent wall, since we define the door relative to the wall's horizontal center.
    """

    thickness: float = 0.02
    """
    Thickness of the door, since the door dictionary only provides a 2d polygon.
    """

    name: PrefixedName = field(init=False)
    """
    Name of the door, constructed from the assetId and room numbers.
    """

    min_x: float = field(init=False)
    """
    Minimum x-coordinate of the door's hole polygon.
    """

    min_y: float = field(init=False)
    """
    Minimum y-coordinate of the door's hole polygon.
    """

    max_x: float = field(init=False)
    """
    Maximum x-coordinate of the door's hole polygon.
    """

    max_y: float = field(init=False)
    """
    Maximum y-coordinate of the door's hole polygon.
    """

    def __post_init__(self):
        """
        Extracts the hole polygon, and preprocesses the name and min/max coordinates of the door's hole polygon.
        """
        asset_id = self.door_dict["assetId"]
        room_numbers = self.door_dict["id"].split("|")[1:]

        self.name = PrefixedName(
            f"{asset_id}_room{room_numbers[0]}_room{room_numbers[1]}"
        )

        hole_polygon = self.door_dict["holePolygon"]

        x0, y0 = float(hole_polygon[0]["x"]), float(hole_polygon[0]["y"])
        x1, y1 = float(hole_polygon[1]["x"]), float(hole_polygon[1]["y"])

        self.x_min, self.x_max = (x0, x1) if x0 <= x1 else (x1, x0)
        self.y_min, self.y_max = (y0, y1) if y0 <= y1 else (y1, y0)

    @cached_property
    def scale(self) -> Scale:
        """
        Computes the door scale from the door's hole polygon. Converts the scale from Unity's left-handed Y-up, Z-forward
        convention to the semantic digital twin's right-handed Z-up, X-forward convention.

        :return: Scale representing the door's geometry.
        """
        width = self.x_max - self.x_min
        height = self.y_max - self.y_min
        return Scale(self.thickness, width, height)

    @cached_property
    def wall_T_door(self) -> TransformationMatrix:
        """
        Computes the door position from the wall's horizontal center. Converts the Unity's left-handed Y-up, Z-forward
        convention to the semantic digital twin's right-handed Z-up, X-forward convention.

        :return: TransformationMatrix representing the door's transform from the wall's perspective.
        """
        # Door center origin expressed from the wall's horizontal center. Unity's wall origin is in one of the corners
        width_origin_wall_corner = 0.5 * (self.x_min + self.x_max)
        height_origin_center = 0.5 * (self.y_min + self.y_max)
        width_origin_center = width_origin_wall_corner - 0.5 * self.parent_wall_width

        # In unity, doors are defined as holes in the wall, so we express them as children of walls.
        # This means we just need to translate them, and can assume no rotation
        return TransformationMatrix.from_point_rotation_matrix(
            Point3(0, -width_origin_center, height_origin_center)
        )

    def _get_double_door_factory(self) -> DoubleDoorFactory:
        """
        Parses the parameters according to the double door assumptions, and returns a double door factory.
        """
        one_door_scale = Scale(self.thickness, self.scale.y * 0.5, self.scale.z)
        x_direction: float = one_door_scale.x / 2
        y_direction: float = one_door_scale.y / 2
        handle_directions = [Direction.Y, Direction.NEGATIVE_Y]

        door_factories = []
        door_transforms = []

        for index, direction in enumerate(handle_directions):
            single_door_name = PrefixedName(
                f"{self.name.name}_{index}", self.name.prefix
            )

            horizontal_direction = (
                HorizontalSemanticDirection.RIGHT
                if direction == Direction.Y
                else HorizontalSemanticDirection.LEFT
            )
            semantic_position = SemanticPositionDescription(
                horizontal_direction_chain=[
                    horizontal_direction,
                    HorizontalSemanticDirection.FULLY_CENTER,
                ],
                vertical_direction_chain=[VerticalSemanticDirection.FULLY_CENTER],
            )

            door_factory = self._get_single_door_factory(
                semantic_position, single_door_name, one_door_scale
            )

            door_factories.append(door_factory)

            parent_T_door = TransformationMatrix.from_xyz_rpy(
                x=x_direction,
                y=(-y_direction) if direction == Direction.Y else y_direction,
            )
            door_transforms.append(parent_T_door)

        double_door_factory = DoubleDoorFactory(
            name=self.name,
            door_factories=door_factories,
            door_transforms=door_transforms,
        )
        return double_door_factory

    def _get_single_door_factory(
        self,
        semantic_position: SemanticPositionDescription,
        name: Optional[PrefixedName] = None,
        scale: Optional[Scale] = None,
    ) -> DoorFactory:
        """
        Parses the parameters according to the single door assumptions, and returns a single door factory.
        """
        name = self.name if name is None else name
        scale = self.scale if scale is None else scale
        handle_name = PrefixedName(f"{name.name}_handle", name.prefix)
        door_factory = DoorFactory(
            name=name,
            scale=scale,
            handle_factory=HandleFactory(name=handle_name),
            semantic_position=semantic_position,
        )
        return door_factory

    def get_factory(self) -> Union[DoorFactory, DoubleDoorFactory]:
        """
        Returns a Factory for the door, either a DoorFactory or a DoubleDoorFactory,
        depending on its name. If the door's name contains "double", it is treated as a double door.
        """

        if "double" in self.name.name.lower():
            return self._get_double_door_factory()
        else:
            semantic_position = SemanticPositionDescription(
                horizontal_direction_chain=[
                    HorizontalSemanticDirection.RIGHT,
                    HorizontalSemanticDirection.FULLY_CENTER,
                ],
                vertical_direction_chain=[VerticalSemanticDirection.FULLY_CENTER],
            )
            return self._get_single_door_factory(semantic_position=semantic_position)


@dataclass
class ProcthorWall:
    """
    Processes a wall dictionary from Procthor, extracting the wall's polygon and computing its scale and
    transformation matrix. Its center will be at the horizontal center of its polygon, at height 0.
     It also processes any doors associated with the wall, creating ProcthorDoor instances for each door.
    The wall is defined by two polygons, one for each side of the physical wall, and the door is defined as a hole in
    the wall polygon.
    """

    wall_dicts: List[dict] = field(default_factory=list)
    """
    List of dictionaries, where each dictionary represents one wall polygon in procthor
    """

    door_dicts: List[dict] = field(default_factory=list)
    """
    List of dictionaries, where each dictionary represents one door hole in the wall polygon
    """

    wall_thickness: float = 0.02
    """
    Thickness of the wall, since the wall dictionary only provides a 2d polygon.
    """

    name: PrefixedName = field(init=False)
    """
    Name of the wall, constructed from the corners of the wall polygon and the room numbers associated with the wall.
    """

    x_coords: List[float] = field(init=False)
    """
    List of unique X-coordinates of the wall polygon, extracted in order from the wall dictionary. 
    """

    y_coords: List[float] = field(init=False)
    """
    List of unique Y-coordinates of the wall polygon, extracted in order from the wall dictionary.
    """

    z_coords: List[float] = field(init=False)
    """
    List of unique Z-coordinates of the wall polygon, extracted in order from the wall dictionary.
    """

    delta_x: float = field(init=False)
    """
    Difference between the first and last X-coordinates of the wall polygon.
    """

    delta_z: float = field(init=False)
    """
    Difference between the first and last Z-coordinates of the wall polygon.
    """

    def __post_init__(self):
        """
        Processes the wall polygons and doors, extracting the min/max coordinates and computing the name of the wall.
        If no doors are present, it uses the first wall polygon as the reference for min/max coordinates.
        If doors are present, it uses the wall polygon that corresponds to the first door's 'wall0' reference.
         This is because the door hole is defined relative to that wall polygon and using the other wall would result
         in the hole being on the wrong side of the wall.
        """
        if self.door_dicts:
            used_wall = (
                self.wall_dicts[0]
                if self.wall_dicts[0]["id"] == self.door_dicts[0]["wall0"]
                else self.wall_dicts[1]
            )
        else:
            used_wall = self.wall_dicts[0]

        polygon = used_wall["polygon"]

        def unique_in_order(seq):
            return list(dict.fromkeys(seq))

        self.x_coords = unique_in_order(float(p["x"]) for p in polygon)
        self.y_coords = unique_in_order(float(p["y"]) for p in polygon)
        self.z_coords = unique_in_order(float(p["z"]) for p in polygon)

        # ProcTHOR wall polygons always have exactly four corners, and are perfectly vertical, so we have at
        # most two unique x and two unique z coordinates. If they line up perfectly, we may have only one unique
        # x or z coordinate, which is why we need to access -1 for generality.
        self.delta_x, self.delta_z = (
            self.x_coords[0] - self.x_coords[-1],
            self.z_coords[0] - self.z_coords[-1],
        )

        room_numbers = [w["id"].split("|")[1] for w in self.wall_dicts]
        corners = used_wall["id"].split("|")[2:]
        self.name = PrefixedName(
            f"wall_{corners[0]}_{corners[1]}_{corners[2]}_{corners[3]}_room{room_numbers[0]}_room{room_numbers[1]}"
        )

    @cached_property
    def scale(self) -> Scale:
        """
        Computes the wall scale from the first wall polygon. Converts the scale from Unity's left-handed Y-up, Z-forward
        convention to the semantic digital twin's right-handed Z-up, X-forward convention.

        :return: Scale representing the wall's geometry.
        """
        width = math.hypot(self.delta_x, self.delta_z)
        min_y, max_y = min(self.y_coords), max(self.y_coords)

        height = max_y - min_y

        return Scale(x=self.wall_thickness, y=width, z=height)

    @cached_property
    def world_T_wall(self) -> TransformationMatrix:
        """
        Computes the wall's world position matrix from the wall's x and z coordinates.
        Calculates the yaw angle using the atan2 function based on the wall's width and depth.
        The wall is artificially set to height=0, because
        1. as of now, procthor house floors have the same floor value at 0
        2. Since doors origins are in 3d center, positioning the door correctly at the floor given potentially varying
           wall heights is unnecessarily complex given the assumption stated in 1.
        """

        yaw = math.atan2(self.delta_z, -self.delta_x)
        x_center = (self.x_coords[0] + self.x_coords[-1]) * 0.5
        z_center = (self.z_coords[0] + self.z_coords[-1]) * 0.5

        world_T_wall = TransformationMatrix.from_xyz_rpy(
            x_center, 0, z_center, 0.0, yaw, 0
        )

        return unity_to_semantic_digital_twin_transform(world_T_wall)

    def get_world(self) -> World:
        """
        Returns a World instance with this wall at its root.
        """
        door_factories = []
        list_wall_T_door = []

        for door in self.door_dicts:
            door = ProcthorDoor(door_dict=door, parent_wall_width=self.scale.y)
            door_factories.append(door.get_factory())
            list_wall_T_door.append(door.wall_T_door)

        wall_factory = WallFactory(
            name=self.name,
            scale=self.scale,
            door_factories=door_factories,
            door_transforms=list_wall_T_door,
        )

        return wall_factory.create()


@dataclass
class ProcthorRoom:
    """
    Processes a room dictionary from Procthor, extracting the room's floor polygon and computing its center.
    """

    room_dict: dict
    """
    Dictionary representing a room from Procthor's JSON format.
    """

    name: PrefixedName = field(init=False)
    """
    Name of the room, constructed from the room type and room ID.
    """

    centered_polytope: List[Point3] = field(init=False)
    """
    Polytope representing the room's floor polygon, centered around its local 0, 0, 0 coordinate
    """

    def __post_init__(self):
        """
        Extracts the room's floor polygon, computes its center, and constructs the centered polytope.
        """
        room_polytope = self.room_dict["floorPolygon"]

        polytope_length = len(room_polytope)
        coords = ((v["x"], v["y"], v["z"]) for v in room_polytope)
        x_coords, y_coords, z_coords = zip(*coords)
        self.x_center = sum(x_coords) / polytope_length
        self.y_center = sum(y_coords) / polytope_length
        self.z_center = sum(z_coords) / polytope_length

        self.centered_polytope = [
            Point3(
                v["z"] - self.z_center,
                -(v["x"] - self.x_center),
                v["y"] - self.y_center,
            )
            for v in room_polytope
        ]

        room_id = self.room_dict["id"].split("|")[-1]
        self.name = PrefixedName(f"{self.room_dict['roomType']}_{room_id}")

    @cached_property
    def world_T_room(self) -> TransformationMatrix:
        """
        Computes the room's world transform
        """

        world_P_room = Point3(self.z_center, -self.x_center, self.y_center)

        return TransformationMatrix.from_point_rotation_matrix(world_P_room)

    def get_world(self) -> World:
        """
        Returns a World instance with this room as a Region at its root.
        """

        return RoomFactory(
            name=self.name, floor_polytope=self.centered_polytope
        ).create()


@dataclass
class ProcthorObject:
    """
    Processes an object dictionary from Procthor, extracting the object's position and rotation,
    and computing its world transformation matrix. It also handles the import of child objects recursively.
    """

    object_dict: dict
    """
    Dictionary representing an object from Procthor's JSON format.
    """

    session: Session
    """
    SQLAlchemy session to interact with the database to import objects.
    """

    @cached_property
    def world_T_obj(self) -> TransformationMatrix:
        """
        Computes the object's world transformation matrix from its position and rotation. Converts Unity's
        left-handed Y-up, Z-forward convention to the right-handed Z-up, X-forward convention.
        """
        obj_position = self.object_dict["position"]
        obj_rotation = self.object_dict["rotation"]
        world_T_obj = TransformationMatrix.from_xyz_rpy(
            obj_position["x"],
            obj_position["y"],
            obj_position["z"],
            math.radians(obj_rotation["x"]),
            math.radians(obj_rotation["y"]),
            math.radians(obj_rotation["z"]),
        )

        return TransformationMatrix(
            unity_to_semantic_digital_twin_transform(world_T_obj)
        )

    def get_world(self) -> Optional[World]:
        """
        Returns a World instance with this object at its root, importing it from the database using its assetId.
        If the object has children, they are imported recursively and connected to the parent object.
        If the object cannot be found in the database, it's children are skipped as well.
        """
        asset_id = self.object_dict["assetId"]
        body_world: World = get_world_by_asset_id(self.session, asset_id=asset_id)

        if body_world is None:
            logging.error(
                f"Could not find asset {asset_id} in the database. Skipping object and its children."
            )
            return None

        with body_world.modify_world():

            for child in self.object_dict.get("children", {}):
                child_object = ProcthorObject(child, self.session)
                world_T_child = child_object.world_T_obj
                child_world = child_object.get_world()
                if child_world is None:
                    continue
                obj_T_child = self.world_T_obj.inverse() @ world_T_child
                child_connection = FixedConnection(
                    parent=body_world.root,
                    child=child_world.root,
                    parent_T_connection_expression=obj_T_child,
                )
                body_world.merge_world(
                    child_world, child_connection, handle_duplicates=True
                )

            return body_world


def unity_to_semantic_digital_twin_transform(
    unity_transform_matrix: TransformationMatrix,
) -> TransformationMatrix:
    """
    Convert a left-handed Y-up, Z-forward Unity transform to the right-handed Z-up, X-forward convention used in the
    semantic digital twin.

    :param unity_transform_matrix:  The transformation matrix in Unity coordinates.
    :return: TransformationMatrix in semantic digital twin coordinates.
    """

    unity_transform_matrix = unity_transform_matrix.to_np()

    permutation_matrix = np.array(
        [
            [0, 0, 1],
            [1, 0, 0],
            [0, 1, 0],
        ],
        dtype=float,
    )

    reflection_vector = np.diag([1, -1, 1])
    R = reflection_vector @ permutation_matrix
    conjugation_matrix = np.eye(4)
    conjugation_matrix[:3, :3] = R
    inverse_conjugation_matrix = conjugation_matrix.T

    unity_transform_matrix = np.asarray(unity_transform_matrix, float).reshape(4, 4)

    return TransformationMatrix(
        data=conjugation_matrix @ unity_transform_matrix @ inverse_conjugation_matrix
    )


@dataclass
class ProcTHORParser:
    """
    Parses a Procthor JSON file into a semantic digital twin World.
    """

    file_path: str
    """
    File path to the Procthor JSON file.
    """

    session: Optional[Session] = field(default=None)
    """
    SQLAlchemy session to interact with the database to import objects.
    """

    def parse(self) -> World:
        """
        Parses a JSON file from procthor into a world.
        Room floor areas are constructed from the supplied polygons
        Walls and doors are constructed from the supplied polygons
        Objects are imported from the database
        """
        with open(self.file_path) as f:
            house = json.load(f)
        house_name = self.file_path.split("/")[-1].split(".")[0]

        world = World(name=house_name)
        with world.modify_world():
            world_root = Body(name=PrefixedName(house_name))
            world.add_kinematic_structure_entity(world_root, handle_duplicates=True)

            self.import_rooms(world, house["rooms"])

            if self.session is not None:
                self.import_objects(world, house["objects"])
            else:
                logging.warning("No database session provided, skipping object import.")

            self.import_walls_and_doors(world, house["walls"], house["doors"])

            return world

    @staticmethod
    def import_rooms(world: World, rooms: List[Dict]):
        """
        Imports rooms from the Procthor JSON file into ProcthorRoom instances.

        :param world: The World instance to which the rooms will be added.
        :param rooms: List of room dictionaries from the Procthor JSON file.
        """
        for room in rooms:
            procthor_room = ProcthorRoom(room_dict=room)
            room_world = procthor_room.get_world()
            room_connection = FixedConnection(
                parent=world.root,
                child=room_world.root,
                parent_T_connection_expression=procthor_room.world_T_room,
            )
            world.merge_world(room_world, room_connection, handle_duplicates=True)

    def import_objects(self, world: World, objects: List[Dict]):
        """
        Imports objects from the Procthor JSON file into ProcthorObject instances.

        :param world: The World instance to which the objects will be added.
        :param objects: List of object dictionaries from the Procthor JSON file.
        """
        for obj in objects:
            procthor_object = ProcthorObject(object_dict=obj, session=self.session)
            obj_world = procthor_object.get_world()
            if obj_world is None:
                continue
            obj_connection = FixedConnection(
                parent=world.root,
                child=obj_world.root,
                parent_T_connection_expression=procthor_object.world_T_obj,
            )
            world.merge_world(obj_world, obj_connection, handle_duplicates=True)

    def import_walls_and_doors(
        self, world: World, walls: List[Dict], doors: List[Dict]
    ):
        """
        Imports walls from the Procthor JSON file into ProcthorWall instances.

        :param world: The World instance to which the walls will be added.
        :param walls: List of wall dictionaries from the Procthor JSON file.
        :param doors: List of door dictionaries from the Procthor JSON file.
        """
        procthor_walls = self._build_procthor_walls(walls, doors)

        for procthor_wall in procthor_walls:
            wall_world = procthor_wall.get_world()
            wall_connection = FixedConnection(
                parent=world.root,
                child=wall_world.root,
                parent_T_connection_expression=procthor_wall.world_T_wall,
            )
            world.merge_world(wall_world, wall_connection, handle_duplicates=True)

    @staticmethod
    def _build_procthor_wall_from_polygon(
        walls: List[Dict],
    ) -> List[ProcthorWall]:
        """
        Groups walls by their polygon and creates ProcthorWall instances for each group.

        :param walls: List of walls without doors

        :return: List of ProcthorWall instances, each representing a pair of walls with the same polygon.
        :raises AssertionError: If the number of walls is not even, as we assume that walls are always paired.
        """

        assert len(walls) % 2 == 0, (
            f"Expected an even number of walls, but found {len(walls)}. "
            f"We assumed that this is never the case, this case may need to be handled now."
        )

        def _polygon_key(poly):
            return frozenset((p["x"], p["y"], p["z"]) for p in poly)

        groups = {}
        for wall in walls:
            key = _polygon_key(wall.get("polygon", []))
            groups.setdefault(key, []).append(wall)

        procthor_walls = [
            ProcthorWall(wall_dicts=matched_walls) for matched_walls in groups.values()
        ]

        return procthor_walls

    @staticmethod
    def _build_procthor_wall_from_door(
        walls: List[Dict], doors: List[Dict]
    ) -> Tuple[List[ProcthorWall], Set[str]]:
        """
        Builds ProcthorWall instances from the provided walls and doors, associating each door with its corresponding walls.

        :param walls: List of wall dictionaries
        :param doors: List of door dictionaries

        :returns: Tuple containing a list of ProcthorWall instances and a set of used wall IDs.
        :raises AssertionError: If a door does not have exactly two walls associated with it.
        """
        walls_by_id = {wall["id"]: wall for wall in walls}
        used_wall_ids = set()
        procthor_walls = []

        for door in doors:
            wall_ids = [door.get("wall0"), door.get("wall1")]
            found_walls = []
            for wall_id in wall_ids:
                wall = walls_by_id[wall_id]
                found_walls.append(wall)
                used_wall_ids.add(wall_id)

            assert (
                len(found_walls) == 2
            ), f"Door {door['id']} should have two walls, but found {len(found_walls)}."

            procthor_walls.append(
                ProcthorWall(door_dicts=[door], wall_dicts=found_walls)
            )

        return procthor_walls, used_wall_ids

    def _build_procthor_walls(
        self, walls: List[Dict], doors: List[Dict]
    ) -> List[ProcthorWall]:
        """
        Builds ProcthorWall instances from the provided walls and doors.

        :param doors: List of door dictionaries
        :param walls: List of wall dictionaries

        :returns: List of ProcthorWall
        """

        procthor_walls, used_wall_ids = self._build_procthor_wall_from_door(
            walls, doors
        )
        remaining_walls = [wall for wall in walls if wall["id"] not in used_wall_ids]
        paired_walls = self._build_procthor_wall_from_polygon(remaining_walls)

        procthor_walls.extend(paired_walls)

        return procthor_walls


def get_world_by_asset_id(session: Session, asset_id: str) -> Optional[World]:
    """
    Queries the database for a WorldMapping with the given asset_id provided by the procthor file.
    """
    asset_id = asset_id.lower()
    expr = the(
        entity(
            world := let(type_=WorldMapping),
            world.name == asset_id,
        )
    )
    other_possible_name = "_".join(asset_id.split("_")[:-1])
    expr2 = the(
        entity(
            world := let(type_=WorldMapping),
            world.name == other_possible_name,
        )
    )
    logging.info(f"Querying name: {asset_id}")
    try:
        world_mapping = eql_to_sql(expr, session).evaluate()
    except NoResultFound:
        try:
            logging.info(f"Querying name: {other_possible_name}")
            world_mapping = eql_to_sql(expr2, session).evaluate()
        except NoResultFound:
            world_mapping = None
            logging.warning(
                f"Could not find world with name {asset_id} or {other_possible_name}; Skipping."
            )

    return world_mapping.from_dao() if world_mapping else None
